'''
Provide a series of tools for getting drug mappings, e.g., name to SMILES string,
drug-DDI matrix, drug ndc - rxcui - atc4 codes, drug cid-atc codes.
'''
import json
import wget
import os
import requests
import re
import pdb
import zipfile
import gzip

import networkx as nx
from networkx.readwrite import json_graph
import pandas as pd
from tqdm import tqdm

from ..utils.tabular_utils import read_csv_to_df, read_txt_to_df, read_excel_to_df
from ..utils.check import make_dir_if_not_exist

# DRUG_DDI_URL = 'https://uofi.box.com/shared/static/xdfgrnhzotz6ktyrdsrikrnz1th97fic.csv'
DRUG_DDI_URL = 'https://storage.googleapis.com/pytrial/drug-DDI.csv'
# DRUG_BANK_URL = 'https://uofi.box.com/shared/static/4f3g4dvdfyz5goubazeqfzdi0abcgzwf.csv'
DRUG_BANK_URL = 'https://storage.googleapis.com/pytrial/drugbank_drugs_info.csv'
RXCUI_ATC4_NDC11_URL = 'https://github.com/RyanWangZf/PyTrial/raw/main/resources/rxcui_atc4_ndc11.zip'
NDC_NAME_URL = 'https://github.com/RyanWangZf/PyTrial/raw/main/resources/ndc_name.csv'
NAME_SMILES_URL = 'https://github.com/RyanWangZf/PyTrial/raw/main/resources/drug_smiles.csv'
NAME_ATC_URL ='https://github.com/RyanWangZf/PyTrial/raw/main/resources/atc_drug.csv'
# ATC5_NDC_URL = 'https://uofi.box.com/shared/static/dk07wip4l4hkbolp09e3rz3un4hokocb.zip'
ATC5_NDC_URL = 'https://storage.googleapis.com/pytrial/atc5_ndc.zip'
# ATC_DEF_URL = 'https://uofi.box.com/shared/static/tdz6glo9waf353mwxvqw44l6r43vrfqm.zip'
ATC_DEF_URL = 'https://storage.googleapis.com/pytrial/ATC.csv.zip'
# FDA_NDC_NAME_URL = 'https://uofi.box.com/shared/static/ah6gk3ljaj0uz0cr2yoecmd3so2onih7.zip'
FDA_NDC_NAME_URL = 'https://storage.googleapis.com/pytrial/fda_ndc.zip'

class DrugTransformer:
    '''
    Provide a series of drug-related functions for

    (1) drug name to atc / atc to drug name
    (2) drug name to ndc-11 / ndc-11 to drug name
    (3) drug name to smiles (molecule structure)
    (4) atc to smiles
    (5) ndc-11 to smiles
    (6) ndc-11 to atc / atc to ndc

    To convert ndc-10 ditis to ndc-11, use `convert_ndc10_ndc11`.
    '''
    def __init__(self):
        # initialize using several preprocessed files
        self._build_ndc_rxcui_atc_map()
        self._build_ndc_name_map()
        self._build_name_smiles()
        self._build_atc_name()
        self._build_atc_ndc_smiles()

    def ndc2atc(self, code):
        '''
        Parameters
        ----------
        code: str or list[str]
            NDC-11 codes.

        Returns
        -------
        atc codes: list[str] or list[list[str,None]]
        '''
        df = self.rxcui_atc4_ndc11
        single_input = isinstance(code, str)
        if single_input: code = [code]
        outputs = []
        for c in code:
            df_sub = df[df['ndc11']==c]
            if len(df_sub) == 0:
                outputs.append(None)
            else:
                outputs.append(df_sub['atc4'].tolist())
        if single_input: return outputs[0]
        else: return outputs

    def atc2ndc(self, code):
        '''
        Parameters
        ----------
        code: str or list[str]
            ATC-4 codes.

        Returns
        -------
        NDC-11 codes: list[str] or list[list[str,None]]
        '''
        df = self.rxcui_atc4_ndc11
        single_input = isinstance(code, str)
        if single_input: code = [code]
        outputs = []
        for c in code:
            df_sub = df[df['atc4']==c]
            if len(df_sub) == 0:
                outputs.append(None)
            else:
                outputs.append(df_sub['ndc11'].tolist())
        if single_input: return outputs[0]
        else: return outputs

    def name2ndc(self, name):
        '''
        Parameters
        ----------
        name: str or list[str]
            Drug names.

        Returns
        -------
        ndc codes: list[str] or list[list[str,None]]
        '''
        df = self.ndc11_name
        single_input = isinstance(name, str)
        if single_input: name = [name]
        outputs = []
        for c in name:
            df_sub = df[df['drug']==c.lower()]
            if len(df_sub) == 0:
                outputs.append(None)
            else:
                outputs.append(df_sub['ndc11'].tolist())
        if single_input: return outputs[0]
        else: return outputs

    def ndc2name(self, code):
        '''
        Parameters
        ----------
        code: str or list[str]
            NDC-11 codes.

        Returns
        -------
        name: list[str] or list[list[str,None]]
            Drug names.
        '''
        df = self.ndc11_name
        single_input = isinstance(code, str)
        if single_input: code = [code]
        outputs = []
        for c in code:
            df_sub = df[df['ndc11']==c.lower()]
            if len(df_sub) == 0:
                outputs.append(None)
            else:
                outputs.append(df_sub['drug'].tolist())
        if single_input: return outputs[0]
        else: return outputs

    def atc2name(self, code):
        '''
        Parameters
        ----------
        code: str or list[str]
            ATC4 codes.

        Returns
        -------
        names: list[str] or list[list[str,None]]
            Drug names.
        '''
        df = self.name_atc
        single_input = isinstance(code, str)
        if single_input: code = [code]
        outputs = []
        for c in code:
            df_sub = df[df['atc4']==c]
            if len(df_sub) == 0:
                outputs.append(None)
            else:
                outputs.append(df_sub['drug'].tolist())
        if single_input: return outputs[0]
        else: return outputs

    def name2atc(self, name):
        '''
        Parameters
        ----------
        name: str or list[str]
            Drug names.

        Returns
        -------
        codes: list[str] or list[list[str,None]]
            ATC4 codes.
        '''
        df = self.name_atc
        single_input = isinstance(name, str)
        if single_input: name = [name]
        outputs = []
        for c in name:
            df_sub = df[df['drug']==c.lower()]
            if len(df_sub) == 0:
                outputs.append(None)
            else:
                outputs.append(df_sub['atc4'].tolist())
        if single_input: return outputs[0]
        else: return outputs

    def name2smiles(self, name):
        '''
        Parameters
        ----------
        name: str or list[str]
            Drug names.

        Returns
        -------
        smiles: list[str] or list[list[str,None]]
            Drug molecule structures represented by SMILES.
        '''
        df = self.name_smiles
        single_input = isinstance(name, str)
        if single_input: name = [name]
        outputs = []
        for c in name:
            df_sub = df[df['drug']==c.lower()]
            if len(df_sub) == 0:
                outputs.append(None)
            else:
                outputs.append(df_sub['moldb_smiles'].tolist())

        if single_input: return outputs[0]
        else: return outputs

    def atc2smiles(self, code):
        '''
        Parameters
        ----------
        code: str or list[str]
            ATC4 codes.

        Returns
        -------
        smiles: list[str] or list[list[str,None]]
            Drug molecule structures represented by SMILES.
        '''
        df = self.atc_smiles
        single_input = isinstance(code, str)
        if single_input: code = [code]
        outputs = []
        for c in code:
            df_sub = df[df['atc4']==c]
            if len(df_sub) == 0:
                outputs.append(None)
            else:
                outputs.append(df_sub['moldb_smiles'].tolist())
        if single_input: return outputs[0]
        else: return outputs

    def ndc2smiles(self, code):
        '''
        Parameters
        ----------
        name: str or list[str]
            Drug names.

        Returns
        -------
        smiles: list[str] or list[list[str,None]]
            Drug molecule structures represented by SMILES.
        '''
        '''
        Parameters
        ----------
        code: str or list[str]
            NDC-11 codes.

        Returns
        -------
        smiles: list[str] or list[list[str,None]]
            Drug molecule structures represented by SMILES.
        '''
        df = self.ndc_smiles
        single_input = isinstance(code, str)
        if single_input: code = [code]
        outputs = []
        for c in code:
            df_sub = df[df['ndc11']==c]
            if len(df_sub) == 0:
                outputs.append(None)
            else:
                outputs.append(df_sub['moldb_smiles'].tolist())
        if single_input: return outputs[0]
        else: return outputs

    def _build_ndc_rxcui_atc_map(self):
        # load ndc_rxcui_atc map csv
        filename = './resources/rxcui_atc4_ndc11.csv'
        make_dir_if_not_exist('resources')
        if not os.path.exists('./resources/rxcui_atc4_ndc11.csv'):
            wget.download(RXCUI_ATC4_NDC11_URL, out=filename.replace('.csv','.zip'))
            f = zipfile.ZipFile(filename.replace('.csv','.zip'),'r')
            f.extractall('./resources')
            f.close()
            print(f'Download RXCUI-NDC-ATC4 mapping file to {filename}.')
        self.rxcui_atc4_ndc11 = read_csv_to_df(filename, dtype={'ndc11':str, 'rxcui':str}, low_memory=False)

    def _build_ndc_name_map(self):
        filename = './resources/ndc_name.csv'
        make_dir_if_not_exist('resources')
        if not os.path.exists(filename):
            wget.download(NDC_NAME_URL, out=filename)
            print(f'Download NDC-DrugName mapping file to {filename}.')
        self.ndc11_name = read_csv_to_df(filename, dtype={'ndc11':str, 'DRUG':str}, low_memory=False)
        self.ndc11_name = self.ndc11_name.applymap(lambda x: x.lower())

    def _build_name_smiles(self):
        filename = './resources/drug_smiles.csv'
        make_dir_if_not_exist('resources')
        if not os.path.exists(filename):
            wget.download(NAME_SMILES_URL, out=filename)
            print(f'Download DrugName-SMILES mapping file to {filename}.')
        self.name_smiles = read_csv_to_df(filename, dtype={'DRUG':str, 'moldb_smiles':str}, low_memory=False)
        self.name_smiles['drug'] = self.name_smiles['drug'].apply(lambda x: x.lower())
        self.name_smiles = self.name_smiles.dropna()

    def _build_atc_name(self):
        filename = './resources/atc_drug.csv'
        make_dir_if_not_exist('resources')
        if not os.path.exists(filename):
            wget.download(NAME_ATC_URL, out=filename)
            print(f'Download DrugName-ATC4 mapping file to {filename}.')
        self.name_atc = read_csv_to_df(filename, dtype={'DRUG':str, 'atc4':str}, low_memory=False)
        self.name_atc['drug'] = self.name_atc['drug'].apply(lambda x: x.lower())

    def _build_atc_ndc_smiles(self):
        df_name_atc = self.name_atc
        df_name_smiles = self.name_smiles
        df = df_name_atc.merge(df_name_smiles, on='drug')[['atc4','moldb_smiles']]
        self.atc_smiles = df.drop_duplicates().reset_index(drop=True)
        df_ndc_name = self.ndc11_name
        df = df_ndc_name.merge(df_name_smiles, on='drug').drop_duplicates().reset_index(drop=True)
        self.ndc_smiles = df[['ndc11','moldb_smiles']]


def convert_ndc10_ndc11(code):
    '''
    Covert NDC 10 digits with hyphens to NDC 11 digits w/o hyphens.

    Examples
    --------

    10-Digit Format: https://health.maryland.gov/phpa/OIDEOR/IMMUN/Shared%20Documents/Handout%203%20-%20NDC%20conversion%20to%2011%20digits.pdf
    on Drug Package	10-Digit Format Example	11-Digit Format	11-Digit Format Example	10-Digit NDC Example	11-Digit Conversion of 10-Digit NDC Example
    4-4-2	9999-9999-99	5-4-2	09999-9999-99	0002-7597-01	00002-7597-01
    5-3-2	99999-999-99	5-4-2	99999-0999-99	50242-040-62	50242-0040-62
    5-4-1	99999-9999-9	5-4-2	99999-9999-09	60574-4114-1	60574-4114-01
    '''
    s = code.split('-')
    S = []
    for i, L in enumerate([5,4,2]):

        if len(s[i]) < L:
            S.append('0'*(L-len(s[i])) + s[i])
        else:
            S.append(s[i])

    return ''.join(S)


def download_drug_ddi(output_dir='./datasets'):
    '''
    Download the drug-ddi information file to disk.

    Parameters
    ----------
    output_dir: str
        The output dir.
    '''
    make_dir_if_not_exist(output_dir)
    filename = os.path.join(output_dir, 'drug-DDI.csv')
    wget.download(DRUG_DDI_URL, out=filename)
    print(f'Save drug DDI file to {filename}.')


def download_drug_bank(output_dir='./datasets'):
    '''
    Download the drug-bank information file to disk.
    Contains drugnames and drug SMILES molecule strings.

    Parameters
    ----------
    output_dir: str
        The output dir.
    '''
    make_dir_if_not_exist(output_dir)
    filename = os.path.join(output_dir, 'drugbank.csv')
    wget.download(DRUG_BANK_URL, out=filename)
    print(f'Save drug bank file to {filename}.')

class DrugGraph:
    '''
    Provide tools to get hierarchy of drug by ACT codes.

    From ATC2 - ATC4 (dont include ATC5)
    '''
    def __init__(self, input_dir='./resources'):
        filename = os.path.join(input_dir, 'drug_hierarchy.json')
        if not os.path.exists(filename):
            self.preprocess(input_dir)

        with open(filename, 'r') as f:
            hierarchy = json.loads(f.read())
        self.graph = json_graph.adjacency_graph(hierarchy)
        print('load drug graph from', filename)

    def preprocess(self, dir='./resources'):
        '''
        Process raw data and deposit the graph data to the local disk.
        Search `ndc_atc.csv` under the given directory `dir`. If not,
        download the raw files to the disk and unzip.
        '''
        make_dir_if_not_exist(dir)

        atctree_file = os.path.join(dir, 'ndc_atc.csv')
        if not os.path.exists(atctree_file):
            self._download_ndc_atc_map(dir, atctree_file)

        atcdef_file = os.path.join(dir, 'atc_def.csv')
        if not os.path.exists(atcdef_file):
            self._download_atc_def(dir, atcdef_file)

        fdandc_file = os.path.join(dir, 'fda_ndc_name.csv')
        if not os.path.exists(fdandc_file):
            self._download_fda_ndc_name(dir, fdandc_file)

        df = read_csv_to_df(atctree_file)
        df['ndc'] = df['ndc'].apply(lambda x: convert_ndc10_ndc11(x))
        df = df.dropna(subset=['atc4']).reset_index(drop=True)
        df_atcdef = read_csv_to_df(atcdef_file)
        df_atcdef['atc'] = df_atcdef['class id'].apply(lambda x: x.split('/')[-1])

        # TODO: attach ndc code to its name
        df_ndc = read_csv_to_df(fdandc_file)
        df_ndc['ndc'] = df_ndc['ndcpackagecode'].apply(lambda x: convert_ndc10_ndc11(x))
        df_ndc = df_ndc[['proprietaryname','ndc']]
        nodes_list = []
        for index, row in df_ndc.iterrows():
            nodes_list.append(
                (row['ndc'], {'description':row['proprietaryname']})
            )

        # add ndc code to the leaf code of G
        G = nx.DiGraph()
        G.add_nodes_from(nodes_list)

        atc4 = df['atc4'].unique()
        for code in tqdm(atc4):
            # some drug dont have atc5 code
            ndc_codes = df[df['atc4'] == code]['ndc'].unique()
            edges = list(zip([code]*len(ndc_codes), ndc_codes))
            G.add_edges_from(edges)

        # build atc code graph
        G_atc = self._build_atc_tree(df_atcdef)
        G = nx.compose(G, G_atc)
        res = nx.adjacency_data(G, attrs={'key':'description', 'id':'id'})

        out_filename = os.path.join(dir, 'drug_hierarchy.json')
        with open(out_filename, 'w') as f:
            f.write(json.dumps(res))
        print('done, save the drug hierarchy to', out_filename)

    def _download_fda_ndc_name(self, dir, fdandc_file):
        temp_filename = os.path.join(dir, 'fda_ndc.zip')
        if os.path.exists(temp_filename): os.remove(temp_filename)
        wget.download(FDA_NDC_NAME_URL, out=temp_filename)
        f = zipfile.ZipFile(temp_filename,'r')
        f.extractall(dir)
        f.close()
        temp_textname = os.path.join(dir, 'package.txt')
        temp_textname2 = os.path.join(dir, 'product.txt')
        df_pack = read_txt_to_df(temp_textname, encoding='cp1252')
        df_prod = read_txt_to_df(temp_textname2, encoding='cp1252')
        df = df_pack.merge(df_prod, on='productid')
        df.to_csv(fdandc_file, index=False)
        os.remove(temp_filename)
        os.remove(temp_textname)
        os.remove(temp_textname2)
        print("\n Download raw file to", fdandc_file)

    def _download_ndc_atc_map(self, dir, atctree_file):
        temp_filename = os.path.join(dir, 'ndc_atc.zip')
        if os.path.exists(temp_filename): os.remove(temp_filename)
        wget.download(ATC5_NDC_URL, out=temp_filename)
        f = zipfile.ZipFile(temp_filename,'r')
        f.extractall(dir)
        f.close()
        temp_csvfilename = os.path.join(dir, 'ndc_map 2020_06_17 (atc5 atc4 ingredients).csv')
        os.rename(temp_csvfilename, atctree_file)
        os.remove(temp_filename)
        print("Download raw file to", atctree_file)

    def _download_atc_def(self, dir, atcdef_file):
        temp_filename = os.path.join(dir, 'atc_def.csv.zip')
        if os.path.exists(temp_filename): os.remove(temp_filename)
        wget.download(ATC_DEF_URL, out=temp_filename)
        f = zipfile.ZipFile(temp_filename,'r')
        f.extractall(dir)
        f.close()
        temp_csvfilename = os.path.join(dir, 'ATC.csv')
        os.rename(temp_csvfilename, atcdef_file)
        os.remove(temp_filename)
        print("\n Download raw file to", atcdef_file)

    def _build_atc_tree(self, df_atcdef):
        G = nx.DiGraph()
        df_atc_label = df_atcdef[['atc','preferred label']]
        df_atc_label = df_atc_label[df_atc_label['atc'].map(len) < 7]
        nodes_list = []
        for index, row in df_atc_label.iterrows():
            nodes_list.append(
                (row['atc'], {'description':row['preferred label']})
            )
        nodes_list.append(('root', {'description':'Anatomical Therapeutic Chemical (ATC) Classification'}))
        G.add_nodes_from(nodes_list)
        codes = df_atcdef['atc'].unique()
        codes = pd.Series(codes)
        atc4 = codes[codes.map(len)==5]
        atc3 = codes[codes.map(len)==4]
        atc2 = codes[codes.map(len)==3]
        atc1 = codes[codes.map(len)==1]
        for _, code in tqdm(atc3.iteritems()):
            children = atc4[atc4.apply(lambda x: code == x[:4])]
            edges = list(zip([code]*len(children), children.tolist()))
            G.add_edges_from(edges)
        for _, code in tqdm(atc2.iteritems()):
            children = atc3[atc3.apply(lambda x: code == x[:3])]
            edges = list(zip([code]*len(children), children.tolist()))
            G.add_edges_from(edges)
        for _, code in tqdm(atc1.iteritems()):
            children = atc2[atc2.apply(lambda x: code == x[:1])]
            edges = list(zip([code]*len(children), children.tolist()))
            G.add_edges_from(edges)

        edges = list(zip(['root']*len(atc1), atc1))
        G.add_edges_from(edges)
        return G

# if __name__ == '__main__':
    # download_drug_ddi()
    # download_drug_bank()



'''
Basic site data functions.
'''
import warnings
import dill
from collections import defaultdict
import torch

from torch.utils.data import Dataset, DataLoader
import numpy as np

from ..utils.tabular_utils import HyperTransformer

class SiteBase(Dataset):
    '''
    Base dataset class for trial site features. Subclass it if additional properties and functions
    are required to add for specific tasks. We make use `rdt`: https://docs.sdv.dev/rdt for transform
    and reverse transform of the tabular data.

    Parameters
    ----------
    df: pd.DataFrame
        The input site tabular format records.

    metadata: dict
        Contains the meta setups of the input data. It should contain the following keys:

        (1) `sdtypes`: dict, the data types of each column in the input data. The keys are the column
            names and the values are the data types. The data types can be one of the following:
            'numerical', 'categorical', 'datetime', 'boolean'.

        (2) `transformers`: dict, the transformers to be used for each column. The keys are the column
            names and the values are the transformer names. The transformer names can be one in
            https://docs.sdv.dev/rdt/transformers-glossary/browse-transformers.

        metadata = {

        'sdtypes': {

        'column1': 'numerical',
        'column2': 'boolean',
        'column3': 'datetime',
        'column4': 'categorical',
        'column5': 'categorical',

        },

        transformers':{

        'column1': rdt.transformers.FloatFormatter(missing_value_replacement='mean'),
        'column2': rdt.transformers.BinaryEncoder(missing_value_replacement='mode'),
        'column3': rdt.transformers.UnixTimestampEncoder(missing_value_replacement='mean'),
        'column4': rdt.transformers.FrequencyEncoder(),
        'column5': None, # will not do any thing to this column if no transformer is specified.
        
        }
        
        } 
        
        It is recommend to provide the metadata of the input tabular data.

        - If no metdata is given, the dataset will automatically detect the `dtypes` of columns and build the corresponding `transformers`.
        - If only `sdtypes` are given, the dataset will detect if there are missing `sdtypes` given and build the `transformers` and dtype automatically.

    transform: bool(default=True)
        Whether or not transform raw self.df by hypertransformer.
        If set False, :code:`self.df` will keep as the same as the passed one.

    Examples
    --------
    >>> from pytrial.data.site_data import TabularSiteBase
    >>> df = pd.read_csv('tabular_site.csv', index_col=0)
    >>> # set `transform=True` will replace dataset.df with dataset.df_transformed
    >>> dataset = SiteBase(df, transform=True)
    >>> # transform raw dataframe to numerical tables
    >>> df_transformed = dataset.transform(df)
    >>> # make back transform to the original df
    >>> df_raw = dataset.reverse_transform(df_transformed)
    '''
    def __init__(self, df, metadata=None, transform=True):
        self.df = df
        self.metadata = metadata

        # initialize hypertransformer
        self.ht = HyperTransformer()

        if transform:
            if metadata is None:
                warnings.warn('No metadata provided. Metadata will be automatically '
                            'detected from your data. This process may not be accurate. '
                            'We recommend writing metadata to ensure correct data handling.')
                self.ht.detect_initial_config(df)
                self.metadata = self.ht.get_config()
                self.ht.fit(df)

            else:
                # parse the metadata and update hypertransformer's config
                self._parse_metadata()

            # replace data with the transformed one
            self.df = self.transform(df)

    def __getitem__(self, index):
        # TODO: support better indexing
        '''
        Indexing the dataframe stored in tabular patient dataset.

        Parameters
        ----------
        index: int or list[int]
            Retrieve the corresponding rows in the dataset.
        '''
        if isinstance(index, int):
            return self.df.iloc[index:index+1]
        
        elif isinstance(index, list):
            return self.df.iloc[index]

    def __len__(self):
        return len(self.df)

    def __repr__(self):
        return f'<pytrial.data.site_data.SiteBase object> Tabular site data with {self.df.shape[0]} samples {self.df.shape[1]} features, call `.df` to yield the pd.DataFrame data: \n' + repr(self.df)

    def transform(self, df=None):
        '''
        Transform the input df or the self.df by hypertransformer.
        If transform=True in `__init__`, then you do not need to call this function
        to transform self.df because it was tranformed already.

        Parameters
        ----------
        df: pd.DataFrame
            The dataframe to be transformed by self.ht
        '''

        if df is None:
            return self.ht.transform(self.df)
        else:
            return self.ht.transform(df)

    def reverse_transform(self, df=None):
        '''
        Reverse the input dataframe back to the original format. Return the self.df in the original
        format if `df=None`.

        Parameters
        ----------
        df: pd.DataFrame
            The dataframe to be transformed back to the original format by self.ht.
        '''
        if df is None:
            return self.ht.reverse_transform(self.df)
        else:
            return self.ht.reverse_transform(df)

    def _parse_metadata(self):
        '''
        Parse the passed metadata, cope with the following scnearios:
        (1) only `sdtypes` are given;
        (2) only `transformers` are given;
        (3) only partial `sdtypes` are given;
        (4) only partial `transformers` are given.
        '''
        # parse metadata dict for building the hypertransformer
        metadata = self.metadata
        self.ht.detect_initial_config(self.df, verbose=False)

        if 'transformers' in metadata:
            self.ht.update_transformers(metadata['transformers'])
        
        if 'sdtypes' in metadata:
            self.ht.update_sdtypes(metadata['sdtypes'])

        self.ht.fit(self.df)
        self.metadata.update(self.ht.get_config())
        
class SiteBaseDemographics(Dataset):
    '''
    Base dataset class for trial site features with demographics added in. Subclass it if additional properties and functions
    are required to add for specific tasks. We make use `rdt`: https://docs.sdv.dev/rdt for transform
    and reverse transform of the tabular data.

    Parameters
    ----------
    df: pd.DataFrame
        The input site tabular format records. Must have column 'demographics' which is a list or ndarray of the racial distribution of the site

    metadata: dict
        Contains the meta setups of the input data. It should contain the following keys:

        (1) `sdtypes`: dict, the data types of each column in the input data. The keys are the column
            names and the values are the data types. The data types can be one of the following:
            'numerical', 'categorical', 'datetime', 'boolean'.

        (2) `transformers`: dict, the transformers to be used for each column. The keys are the column
            names and the values are the transformer names. The transformer names can be one in
            https://docs.sdv.dev/rdt/transformers-glossary/browse-transformers.

        metadata = {

        'sdtypes': {

        'column1': 'numerical',
        'column2': 'boolean',
        'column3': 'datetime',
        'column4': 'categorical',
        'column5': 'categorical',

        },

        transformers':{

        'column1': rdt.transformers.FloatFormatter(missing_value_replacement='mean'),
        'column2': rdt.transformers.BinaryEncoder(missing_value_replacement='mode'),
        'column3': rdt.transformers.UnixTimestampEncoder(missing_value_replacement='mean'),
        'column4': rdt.transformers.FrequencyEncoder(),
        'column5': None, # will not do any thing to this column if no transformer is specified.
        
        }
        
        } 
        
        It is recommend to provide the metadata of the input tabular data.

        - If no metdata is given, the dataset will automatically detect the `dtypes` of columns and build the corresponding `transformers`.
        - If only `sdtypes` are given, the dataset will detect if there are missing `sdtypes` given and build the `transformers` and dtype automatically.

    transform: bool(default=True)
        Whether or not transform raw self.df by hypertransformer.
        If set False, :code:`self.df` will keep as the same as the passed one.

    Examples
    --------
    >>> from pytrial.data.site_data import TabularSiteBase
    >>> df = pd.read_csv('tabular_site.csv', index_col=0)
    >>> # set `transform=True` will replace dataset.df with dataset.df_transformed
    >>> dataset = TabularPatientBase(df, transform=True)
    >>> # transform raw dataframe to numerical tables
    >>> df_transformed = dataset.transform(df)
    >>> # make back transform to the original df
    >>> df_raw = dataset.reverse_transform(df_transformed)
    '''
    def __init__(self, df, metadata=None, transform=True):
        self.eth_labels = np.array(df.pop('demographics'))
        self.df = df
        self.metadata = metadata

        # initialize hypertransformer
        self.ht = HyperTransformer()

        if transform:
            if metadata is None:
                warnings.warn('No metadata provided. Metadata will be automatically '
                            'detected from your data. This process may not be accurate. '
                            'We recommend writing metadata to ensure correct data handling.')
                self.ht.detect_initial_config(df)
                self.metadata = self.ht.get_config()
                self.ht.fit(df)

            else:
                # parse the metadata and update hypertransformer's config
                self._parse_metadata()

            # replace data with the transformed one
            self.df = self.transform(df)

    def __getitem__(self, index):
        # TODO: support better indexing
        '''
        Indexing the dataframe stored in tabular patient dataset.

        Parameters
        ----------
        index: int or list[int]
            Retrieve the corresponding rows in the dataset.
        '''
        if isinstance(index, int):
            return self.df.iloc[index:index+1]
        
        elif isinstance(index, list):
            return self.df.iloc[index]
        
    def get_label(self, index):
        '''
        Indexing the demographic labels

        Parameters
        ----------
        index: int or list[int]
            Retrieve the corresponding rows in the data.
        '''
        return self.eth_labels[index]

    def __len__(self):
        return len(self.df)

    def __repr__(self):
        return f'<pytrial.data.site_data.SiteBase object> Tabular site data with {self.df.shape[0]} samples {self.df.shape[1]} features, call `.df` to yield the pd.DataFrame data: \n' + repr(self.df)

    def transform(self, df=None):
        '''
        Transform the input df or the self.df by hypertransformer.
        If transform=True in `__init__`, then you do not need to call this function
        to transform self.df because it was tranformed already.

        Parameters
        ----------
        df: pd.DataFrame
            The dataframe to be transformed by self.ht
        '''

        if df is None:
            return self.ht.transform(self.df)
        else:
            return self.ht.transform(df)

    def reverse_transform(self, df=None):
        '''
        Reverse the input dataframe back to the original format. Return the self.df in the original
        format if `df=None`.

        Parameters
        ----------
        df: pd.DataFrame
            The dataframe to be transformed back to the original format by self.ht.
        '''
        if df is None:
            return self.ht.reverse_transform(self.df)
        else:
            return self.ht.reverse_transform(df)

    def _parse_metadata(self):
        '''
        Parse the passed metadata, cope with the following scnearios:
        (1) only `sdtypes` are given;
        (2) only `transformers` are given;
        (3) only partial `sdtypes` are given;
        (4) only partial `transformers` are given.
        '''
        # parse metadata dict for building the hypertransformer
        metadata = self.metadata
        self.ht.detect_initial_config(self.df, verbose=False)

        if 'transformers' in metadata:
            self.ht.update_transformers(metadata['transformers'])
        
        if 'sdtypes' in metadata:
            self.ht.update_sdtypes(metadata['sdtypes'])

        self.ht.fit(self.df)
        self.metadata.update(self.ht.get_config())


class ModalitySiteBase(Dataset):
    '''
    Load sequential site inputs for longitudinal records generation.

    Parameters
    ----------
    data: dict
        A dict contains site data in sequence and/or in tabular.

        data = {

        'x': np.ndarray or pd.DataFrame
        
        Static site features in tabular form, typically those specialty information.

        'dx': list or np.ndarray
        
        Diagnosis history sequence in dense format or in tensor format (depends on the model input requirement.)
            
        - If in dense format, it is like [[c1,c2,c3],[c4,c5],...], with shape [n_patient, NA, NA];

        - If in tensor format, it is like [[0,1,1],[1,1,0],...] (multi-hot encoded), with shape [n_patient, max_num_visit, max_num_event].
        
        'rx': list or np.ndarray
        
        Prescription history sequence in dense format or in tensor format (depends on the model input requirement.)
            
        - If in dense format, it is like [[c1,c2,c3],[c4,c5],...], with shape [n_patient, NA, NA];

        - If in tensor format, it is like [[0,1,1],[1,1,0],...] (multi-hot encoded), with shape [n_patient, max_num_visit, max_num_event].
        
        'hist': list or np.ndarray
        
        Past trial enrollment sequence in dense format or in tensor format (depends on the model input requirement.)
            
        - If in dense format, it is like [(t1, e1),(t2, e2),...], with shape [n_past_trial, NA, NA];

        - If in tensor format, it is like [[0,1,1 ... e1],[1,1,0 ... e2],...] (multi-hot encoded), with shape [n_patient, max_num_trial, trial_dim + 1].
        
        'eth_label': np.ndarray

        - Distribution of racial demographics of patients in the site (array should sum to 1 or 100)
        
        }
    
    metadata: dict (optional)
        
        A dict contains configuration of input site data.

        metadata = {

        'dx_visit': dict[str]
        
        a dict contains the format of input data for processing input diagnosis visit sequences.

        `dx_visit`: {

        'mode': 'tensor' or 'dense',
        
        'voc': int,
        
        },
        
        'rx_visit': dict[str]
        
        a dict contains the format of input data for processing input prescription visit sequences.

        `rx_visit`: {

        'mode': 'tensor' or 'dense',
        
        'voc': int,
        
        },
        
        `hist`: {

        'mode': 'tensor' or 'dense',
        
        },
    
        'max_visit': int

        the maximum number of visits for the diagnosis and prescription history modalities considered when building tensor inputs, ignored
        when visit mode is dense.
        
        'max_trial': int

        the maximum number of past trials for the enrollment history modality considered when building tensor inputs, ignored
        when visit mode is dense.
        
        }

    '''
    feature = None
    dx_hist = None
    rx_hist = None
    enroll_hist = None
    eth_label = None
    max_visit = None
    max_trial = None
    visit_voc_size = None

    metadata = {
        'voc': {
            'dx': 100,
            'rx': 100
            },
        
        'feature':{
            'mode': 'tensor',
            },
        
        'visit':{
            'mode': 'tensor',
            },
        
        'hist':{
            'mode': 'tensor',
            },

        'max_visit': 20,
        'max_trial': 10,
    }

    def __init__(self, data, metadata=None) -> None:
        # parse metadata
        self._parse_metadata(metadata=metadata)

        # get input data
        self._parse_inputdata(inputs=data)

    def __getitem__(self, index):
        return_data = {}
        if self.dx_hist is not None:
            dx_hist = self.dx_hist[index]
            if self.metadata['visit']['mode'] == 'tensor':
                dx_ts, dx_len = self._dense_visit_to_tensor(dx_hist, self.metadata['voc']['dx'])
                return_data['dx'] = dx_ts
                return_data['dx_lens'] = dx_len
            else:
                visit_dict, dx_len = self._tensor_visit_to_dense(dx_hist)
                return_data['dx'] = visit_dict
                return_data['dx_lens'] = dx_len
                
        if self.rx_hist is not None:
            rx_hist = self.rx_hist[index]
            if self.metadata['visit']['mode'] == 'tensor':
                rx_ts, rx_len = self._dense_visit_to_tensor(rx_hist, self.metadata['voc']['rx'])
                return_data['rx'] = rx_ts
                return_data['rx_lens'] = rx_len
            else:
                visit_dict, rx_len = self._tensor_visit_to_dense(rx_hist)
                return_data['rx'] = visit_dict
                return_data['rx_lens'] = rx_len
                
        if self.enroll_hist is not None:
            enroll_hist = self.enroll_hist[index]
            if self.metadata['hist']['mode'] == 'tensor':
                enroll_ts, enroll_len = self._dense_trial_to_tensor(enroll_hist) # return a dict with keys corresponding to order
                return_data['hist'] = enroll_ts
                return_data['hist_lens'] = enroll_len
            else:
                enroll_list, enroll_len = self._tensor_trial_to_dense(enroll_hist)
                return_data['hist'] = enroll_list
                return_data['hist_lens'] = enroll_len
                
        if self.feature is not None:
            return_data['x'] = self.feature[index]
            
        if self.eth_label is not None:
            return_data['eth_label'] = self.eth_label[index]
        
        return return_data

    def __len__(self):
        return len(self.visit)

    def _read_pickle(self, file_loc):
        return dill.load(open(file_loc, 'rb'))
    
    def _parse_metadata(self, metadata):
        if metadata is not None: 
            for k,v in metadata.items():
                if isinstance(v, dict):
                    self.metadata[k].update(v)
                else:
                    self.metadata[k] = v
        metadata = self.metadata
        
        if 'max_visit' in metadata:
            self.max_visit = metadata['max_visit']
            
        if 'max_trial' in metadata:
            self.max_trial = metadata['max_trial']

    def _parse_inputdata(self, inputs):
        if 'x' in inputs: self.feature = inputs['x']
        if 'dx' in inputs: self.dx_hist = inputs['dx']
        if 'rx' in inputs: self.rx_hist = inputs['rx']
        if 'hist' in inputs: self.enroll_hist = inputs['hist']
        if 'eth_label' in inputs: self.eth_label = inputs['eth_label']

    def _dense_visit_to_tensor(self, visits, max_size):
        if not isinstance(visits, list):
            numVisits = self.max_visit
            for i in range(self.max_visit):
                if visits[i].sum() == 0:
                    numVisits = i
                    break
            return visits, numVisits
        
        res = np.zeros((self.max_visit, max_size), dtype=int)
        for i, visit in enumerate(visits):
            # clip if the max visit is larger than self.max_visit
            if i >= self.max_visit: break
            res[i, visit] = 1
        return res, min(len(visits), self.max_visit)
    
    def _tensor_visit_to_dense(self, visits):
        if isinstance(visits, list):
            return visits, len(visits)
        
        res = []
        for i in range(len(visits)):
            # clip if the max visit is larger than self.max_visit
            if i >= self.max_visit: break
            codes = np.nonzero(visits[i])[0].tolist()
            if codes == []: break
            res.append(codes)
        return res, min(len(res), self.max_visit)
    
    def _dense_trial_to_tensor(self, trials):
        if not isinstance(trials, list):
            numTrials = self.max_trial
            for i in range(self.max_visit):
                if trials[i].sum() == 0:
                    numTrials = i
                    break
            return trials, numTrials
        
        max_size = 1
        for i in range(len(trials)):
            max_size = trials[i][0]
            break
        
        res = np.zeros((self.max_trial, max_size), dtype=int)
        for i, trial in enumerate(trials):
            # clip if the max trial is larger than self.max_trial
            if i >= self.max_trial: break
            res[i, :-1] = trial[0]
            res[i, -1:] = trial[1]
        return res, min(len(trial), self.max_trial)
    
    def _tensor_trial_to_dense(self, trials):
        if isinstance(trials, list):
            return trials, len(trials)
        
        res = []
        for i in range(len(trials)):
            # clip if the max trial is larger than self.max_trial
            if i >= self.max_trial: break
            if trials[i].sum() == 0: break
            res.append((trials[i][:-1], trials[i][-1]))
        return res, min(len(res), self.max_visit)
    
    def get_label(self):
        return self.eth_label


'''
Provide an easy-to-access to all pretrained BERT-like models based on transformers@huggingface
cuz many models use pretrained bert embeddings.
'''
import pdb

import numpy as np
import torch
from torch import nn
from transformers import AutoModel, AutoTokenizer
from tqdm.autonotebook import trange


class BERT(nn.Module):
    '''
    The pretrained BERT model for getting text embeddings.

    Parameters
    ----------
    bertname: str (default='emilyalsentzer/Bio_ClinicalBERT')
        The name of pretrained bert to get from huggingface models hub: https://huggingface.co/models.
        Or pass the dir where the local pretrained bert is available.

    proj_dim: int or None
        A linear projection head added on top of the bert encoder. Note that if given,
        the projection head is RANDOMLY initialized and needs further training.

    max_length: int
        Maximum acceptable number of tokens for each sentence.

    device: str
        The device of this model, typically be 'cpu' or 'cuda:0'.

    Examples
    --------
    >>> model = BERT()
    >>> emb = model.encode('The goal of life is comfort.')
    >>> print(emb.shape)
    '''
    is_train=None
    def __init__(self, bertname='emilyalsentzer/Bio_ClinicalBERT', proj_dim=None, max_length=512, device='cpu'):
        super().__init__()
        self.projection_head = None
        self.model = AutoModel.from_pretrained(bertname, output_hidden_states=True)
        self.tokenizer = AutoTokenizer.from_pretrained(bertname)
        self.tokenizer.__dict__['model_max_length'] = max_length
        if proj_dim is not None:
            self.projection_head = nn.Linear(768, proj_dim, bias=False)
            self.projection_head.to(device)
        self.device = device
        self.model.to(device)

    def forward(self, input_ids, attention_mask=None, token_type_ids=None, return_hidden_states=False):
        '''
        Forward pass of the model. 
        
        Parameters
        ----------
        input_ids: torch.Tensor
            The input token ids with shape [batch_size, seq_len].
        
        attention_mask: torch.Tensor
            The attention mask with shape [batch_size, seq_len].
        
        token_type_ids: torch.Tensor
            The token type ids with shape [batch_size, seq_len].
        
        return_hidden_states: bool
            Whether to return the hidden states of all layers.
        '''
        output = self.model(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids, return_dict=True)
        if not return_hidden_states:
            embed = output['pooler_output']
            if self.projection_head is not None:
                embed = self.projection_head(embed)
            return embed
        else:
            hidden_states = output['hidden_states'] # input embeds+12 layers, 13 embeds in total
            return hidden_states

    def encode(self, input_text, is_train=False, batch_size=None):
        '''
        Encode the input texts into embeddings.

        Parameters
        ----------
        input_text: str or list[str]
            A sentence or a list of sentences to be encoded.

        is_train: bool
            Set True if this model's parameters will update by learning.

        batch_size: int
            How large batch size to use when encoding long documents with many sentences.
            When set `None`, will encode all sentences at once.

        Returns
        -------
        outputs: torch.Tensor
            The encoded sentence embeddings with size [num_sent, emb_dim]
        '''
        self.is_train = is_train
        if batch_size is not None:
            # smart batching
            all_embeddings = []
            sentences_sorted, length_sorted_idx = self._smart_batching(input_text)
            for start_index in trange(0, len(input_text), batch_size, desc=f'BERT encoding total samples {len(input_text)}'):
                sentences_batch = sentences_sorted[start_index:start_index+batch_size]
                embs = self._encode_batch(sentences_batch)
                all_embeddings.extend(embs)
            
            all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]
            return torch.stack(all_embeddings)
        else:
            return self._encode_batch(input_text)

    def _encode_batch(self, input_text):
        inputs = self.tokenizer(input_text, truncation=True, padding=True, return_tensors='pt')
        inputs = self._to_device(inputs)
        if not self.is_train:
            with torch.no_grad():
                outputs = self.forward(**inputs)
        else:
            outputs = self.forward(**inputs)
        return outputs
    
    def _smart_batching(self, input_text):
        length_sorted_idx = np.argsort([-len(sen) for sen in input_text])
        sentences_sorted = [input_text[idx] for idx in length_sorted_idx]
        return sentences_sorted, length_sorted_idx

    def _to_device(self, inputs):
        for k,v in inputs.items():
            if isinstance(v, torch.Tensor):
                inputs[k] = v.to(self.device)
        return inputs

if __name__ == '__main__':
    bert = BERT()
    emb = bert.encode(['asdadasdas', 'asdasdawqeqw'])
    print(emb.shape)



'''
Several utilities for processing tabular datasets.

TODO:

add LabelEncoder, OneHotEncoder, FrequencyEncoder (all encoders for categorical features) that supports

max_categories and min_frequencies, refer to https://scikit-learn.org/stable/modules/preprocessing.html.
'''
from collections import defaultdict
import pickle
from pathlib import Path
import pdb
import os
from copy import deepcopy

import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler as sk_standardscaler
from sklearn.preprocessing import MinMaxScaler as sk_minmaxscaler
import rdt
from rdt.hyper_transformer import Config
from rdt.errors import InvalidConfigError
from rdt.transformers import BaseTransformer
from rdt.transformers import LabelEncoder, BinaryEncoder, UnixTimestampEncoder
from rdt.transformers.numerical import FloatFormatter
from rdt.transformers.categorical import OneHotEncoder, FrequencyEncoder

class StandardScaler(FloatFormatter):
    '''Transformer for numerical data.

    This transformer scales all numerical values within the same column using
    `sklearn.preprocessing.StandardScaler`.
    
    Null values are replaced using a `NullTransformer` from rdt.

    Parameters
    ----------
    missing_value_replacement: object or None
        Indicate what to do with the null values. If an integer or float is given,
        replace them with the given value. If the strings ``'mean'`` or ``'mode'`` are
        given, replace them with the corresponding aggregation. If ``None`` is given,
        do not replace them. Defaults to ``None``.

    enforce_min_max_values: bool (default=False)
        Whether or not to clip the data returned by ``reverse_transform`` to the min and
        max values seen during ``fit``. Defaults to ``False``.

    learn_rounding_scheme (bool):
        Whether or not to learn what place to round to based on the data seen during ``fit``.
        If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
        Defaults to ``False``.

    missing_value_generation (str or None):
        The way missing values are being handled. There are three strategies:

            * ``random``: Randomly generates missing values based on the percentage of
                missing values.
            * ``from_column``: Creates a binary column that describes whether the original
                value was missing. Then use it to recreate missing values.
            * ``None``: Do nothing with the missing values on the reverse transform. Simply
                pass whatever data we get through.

    computer_representation (dtype):
        Accepts ``'Int8'``, ``'Int16'``, ``'Int32'``, ``'Int64'``, ``'UInt8'``, ``'UInt16'``,
        ``'UInt32'``, ``'UInt64'``, ``'Float'``.
        Defaults to ``'Float'``.

    model_missing_values (bool):
        **Deprecated** Whether to create a new column to indicate which values were null or not. The column
        will be created only if there are null values. If ``True``, create the new column if
        there are null values. If ``False``, do not create the new column even if there
        are null values. Defaults to ``False``.
    '''
    _dtype = None
    _min_value = None
    _max_value = None
    random_states = None

    def __init__(self,
        missing_value_replacement=None,
        enforce_min_max_values=False,
        learn_rounding_scheme=False,
        computer_representation='Float',
        missing_value_generation="random",
        model_missing_values=None,
        ):
        self.output_properties = {None: {'sdtype': 'float', 'next_transformer': None}}
        super().__init__(
            missing_value_replacement=missing_value_replacement,
            model_missing_values=model_missing_values,
            learn_rounding_scheme=learn_rounding_scheme,
            enforce_min_max_values=enforce_min_max_values,
            computer_representation=computer_representation,
            missing_value_generation=missing_value_generation,
        )
        self.standard_transformer = sk_standardscaler()

    def _fit(self, data):
        '''
        Fit the transformer to the data.

        Parameters
        ----------
        data: pd.Series
            Data to fit.
        '''
        super()._fit(data)
        data = super()._transform(data)
        self.standard_transformer.fit(data[:,None])

    def _transform(self, data):
        '''
        Transform numerical data.

        Parameters
        ----------
        data: pd.Series
            Data to transform.

        Returns
        -------
            np.ndarray
        '''
        transformed = super()._transform(data)
        transformed = self.standard_transformer.transform(transformed[:,None])
        return transformed.flatten()

    def _reverse_transform(self, data):
        '''
        Convert the transformed data back to the original format.

        Parameters
        ----------
            data: pd.Series or np.ndarray
                Data to be reversely transformed.

        Returns
        -------
            np.ndarray
        '''
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        if self.missing_value_replacement is not None:
            data = self.null_transformer.reverse_transform(data)

        if self.enforce_min_max_values:
            data = data.clip(self._min_value, self._max_value)

        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        data = self.standard_transformer.inverse_transform(data[:,None])

        is_integer = np.dtype(self._dtype).kind == 'i'
        if self.learn_rounding_scheme or is_integer:
            data = data.round(self._rounding_digits or 0)

        if pd.isna(data).any() and is_integer:
            return data

        return data.astype(self._dtype)

class MinMaxScaler(StandardScaler):
    '''Transformer for numerical data.

    This transformer scales all numerical values within the same column using
    `sklearn.preprocessing.MinMaxScaler`.

    Null values are replaced using a `NullTransformer` from rdt.

    Parameters
    ----------
    missing_value_replacement: object or None
        Indicate what to do with the null values. If an integer or float is given,
        replace them with the given value. If the strings ``'mean'`` or ``'mode'`` are
        given, replace them with the corresponding aggregation. If ``None`` is given,
        do not replace them. Defaults to ``None``.

    enforce_min_max_values: bool (default=False)
        Whether or not to clip the data returned by ``reverse_transform`` to the min and
        max values seen during ``fit``. Defaults to ``False``.

    learn_rounding_scheme (bool):
        Whether or not to learn what place to round to based on the data seen during ``fit``.
        If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
        Defaults to ``False``.

    model_missing_values (bool):
        Whether to create a new column to indicate which values were null or not. The column
        will be created only if there are null values. If ``True``, create the new column if
        there are null values. If ``False``, do not create the new column even if there
        are null values. Defaults to ``False``.
    '''

    _dtype = None
    _min_value = None
    _max_value = None

    def __init__(self,
        missing_value_replacement=None,
        enforce_min_max_values=False,
        learn_rounding_scheme=False,
        model_missing_values=None,
        computer_representation='Float',
        ):
        super().__init__(
            missing_value_replacement=missing_value_replacement,
            enforce_min_max_values=enforce_min_max_values,
            learn_rounding_scheme=learn_rounding_scheme,
            model_missing_values=model_missing_values,
            computer_representation=computer_representation)

        self.standard_transformer = sk_minmaxscaler()


class HyperTransformer(rdt.HyperTransformer):
    '''
    A subclass of `rdt.HyperTransformer` to set special setups.
    '''
    _DTYPES_TO_SDTYPES = {
        'i': 'categorical', # change the default from numerical to categorical for integers
        'f': 'numerical',
        'O': 'categorical',
        'b': 'boolean',
        'M': 'datetime',
    }
    def __init__(self):
        self._default_sdtype_transformers = {
            'numerical': StandardScaler(missing_value_replacement='mean'),
            'categorical': LabelEncoder(),
            'boolean': BinaryEncoder(missing_value_replacement='mode'),
            'datetime': UnixTimestampEncoder(missing_value_replacement='mean'),
        }
        self.field_sdtypes = {}
        self.field_transformers = {}
        self._specified_fields = set()
        self._validate_field_transformers()
        self._valid_output_sdtypes = self._DEFAULT_OUTPUT_SDTYPES
        self._multi_column_fields = self._create_multi_column_fields()
        self._transformers_sequence = []
        self._output_columns = []
        self._input_columns = []
        self._fitted_fields = set()
        self._fitted = False
        self._modified_config = False
        self._transformers_tree = defaultdict(dict)
        self.computer_representation = None

    def detect_initial_config(self, data, verbose=False):
        """Print the configuration of the data.
        This method detects the ``sdtype`` and transformer of each field in the data
        and then prints them as a json object.
        NOTE: This method completely resets the state of the ``HyperTransformer``.
        Args:
            data (pd.DataFrame):
                Data which will have its configuration detected.

            verbose (bool):
                Whether print user message or not.
        """
        # Reset the state of the HyperTransformer
        self.field_sdtypes = {}
        self.field_transformers = {}

        # Set the sdtypes and transformers of all fields to their defaults
        self._learn_config(data)

        if verbose:
            self._user_message('Detecting a new config from the data ... SUCCESS')
            self._user_message('Setting the new config ... SUCCESS')

        config = Config({
            'sdtypes': self.field_sdtypes,
            'transformers': self.field_transformers
        })

        if verbose:
            self._user_message('Config:')
            self._user_message(config)

    @staticmethod
    def _validate_transformers(column_name_to_transformer):
        """Validate the given transformers are valid.
        Args:
            column_name_to_transformer (dict):
                Dict mapping column names to transformers to be used for that column.
        Raises:
            Error:
                Raises an error if ``column_name_to_transformer`` contains one or more
                invalid transformers.
        """
        invalid_transformers_columns = []
        update_transformers = {}
        for column_name, transformer in column_name_to_transformer.items():
            if transformer and not isinstance(transformer, BaseTransformer):
                if isinstance(transformer, str):
                    if get_transformer(transformer) is None:
                        invalid_transformers_columns.append(column_name)
                    else:
                        # update the column_name_to_transformer dict with str inputs
                        update_transformers[column_name] = get_transformer(transformer)()

        if invalid_transformers_columns:
            raise InvalidConfigError(
                f'Invalid transformers for columns: {invalid_transformers_columns}. '
                'Please assign an rdt transformer instance to each column name.'
            )
        
        column_name_to_transformer.update(update_transformers)
    
    def _set_field_sdtype(self, data, field):
        clean_data = data[field].dropna()
        kind = clean_data.infer_objects().dtype.kind
        if kind == 'i':
            # decide if it is categorical or binary for input integers
            if len(clean_data.unique()) <= 2:
                kind = 'b'
            else:
                kind = 'O'
        self.field_sdtypes[field] = self._DTYPES_TO_SDTYPES[kind]


def read_csv_to_df(file_loc, header_lower=True, usecols=None, dtype=None,
                   low_memory=True, encoding=None, index_col=None):
    """Read in csv files with necessary processing
    Parameters
    ----------
    file_loc
    header_lower
    low_memory
    Returns
    -------
    """
    if dtype != None:
        df = pd.read_csv(file_loc, usecols=usecols, dtype=dtype,
                         low_memory=low_memory, encoding=encoding, index_col=index_col)
    else:
        df = pd.read_csv(file_loc, usecols=usecols, low_memory=low_memory,
                         encoding=encoding, index_col=index_col)

    if header_lower:
        df.columns = df.columns.str.lower()
    return df

def read_excel_to_df(file_loc, header_lower=True, usecols=None, dtype=None):
    """Read in excel files with necessary processing
    Parameters
    ----------
    file_loc
    header_lower
    low_memory
    Returns
    -------
    """
    if dtype != None:
        df = pd.read_excel(file_loc, usecols=usecols, dtype=dtype)
    else:
        df = pd.read_excel(file_loc, usecols=usecols)
    if header_lower:
        df.columns = df.columns.str.lower()
    return df

def read_txt_to_df(file_loc, header_lower=True, usecols=None, dtype=None,
                     low_memory=True, encoding=None):
    """Read in excel files with necessary processing

    Parameters
    ----------
    file_loc
    header_lower
    low_memory

    """
    if dtype != None:
        df = pd.read_table(file_loc, usecols=usecols, dtype=dtype,
                           low_memory=low_memory, encoding=encoding)
    else:
        df = pd.read_table(file_loc, usecols=usecols, low_memory=low_memory,
                           encoding=encoding)

    if header_lower:
        df.columns = df.columns.str.lower()
    return df

def read_text_to_list(file_loc, encoding='utf-8'):
    '''
    Read raw text files into a list one line by one line
    '''
    with open(file_loc, 'r', encoding=encoding) as f:
        x = [l.strip() for l in f.readlines()]
    return x


def load_table_config(data_dir, discriminate_bin_feat=False):
    '''
    Load the categorical, numerical, binary feature configuration from the local dataset folder.

    Parameters
    ----------
    data_dir: str
        The target dataset folder.

    discriminate_bin_feat: bool
        Whether discriminate binary and categorical feature or not.

    '''
    num_filepath = os.path.join(data_dir, 'numerical_feature.txt')
    if os.path.exists(num_filepath):
        num_feat_list = read_text_to_list(num_filepath)
    else:
        num_feat_list = []
    
    bin_filepath = os.path.join(data_dir, 'binary_feature.txt')
    if os.path.exists(bin_filepath):
        bin_feat_list = read_text_to_list(bin_filepath)
    else:
        bin_feat_list = []
    bin_feat_list = [c.strip().lower() for c in bin_feat_list]

    # get all column names
    data_filepath = os.path.join(data_dir, 'data_processed.csv')
    df = read_csv_to_df(data_filepath, index_col=0)

    all_feat_list = [c.strip().lower() for c in df.columns.tolist()]
    cat_feat_list = [c for c in all_feat_list if c not in num_feat_list and c not in bin_feat_list and 'target_label' not in c]
    num_feat_list = [c.strip().lower() for c in num_feat_list]

    if not discriminate_bin_feat:
        # merge bin and cat
        cat_feat_list.extend(bin_feat_list)

    # get cat cardinalities
    cat_cardinalities = []
    for c in cat_feat_list:
        cat_cardinalities.append(df[c].nunique())

    return {
        'columns': all_feat_list,
        'num_feat': num_feat_list,
        'bin_feat': bin_feat_list,
        'cat_feat': cat_feat_list,
        'cat_cardinalities': cat_cardinalities,
    }

def get_transformer(name):
    mapping = {
        'labelencoder':LabelEncoder,
        'onehotencoder':OneHotEncoder,
        'binaryencoder':BinaryEncoder,
        'floatformatter':FloatFormatter,
        'frequencyencoder':FrequencyEncoder,
        'unixtimeformatter':UnixTimestampEncoder,
    }
    if isinstance(name, str):
        return mapping.get(name.lower(), None)
    else:
        raise ValueError('The input transformer name must be a string. Get {} instead.'.format(type(name)))


import pdb

from torch.utils.data import DataLoader

from pytrial.utils.trainer import Trainer
from .metrics import METRICS_DICT

class IndivTabTrainer(Trainer):
    '''
    Subclass the original trainer and provide specific evaluation functions and test_dataloader.
    '''
    def evaluate(self):
        mode = self.model.config['mode']
        metric_fn = METRICS_DICT[mode]
        res = self.model._predict_on_dataloader(self.test_dataloader)
        eval_res = metric_fn(res['pred'], res['label'])
        return eval_res

    def get_test_dataloader(self, test_data):
        dataset = self.model._build_dataset(test_data)
        dataloader = DataLoader(dataset, 
            batch_size=self.model.config['batch_size'], 
            shuffle=False, 
            num_workers=self.model.config['num_worker'],
            pin_memory=True,
            )        
        return dataloader

class IndivSeqTrainer(Trainer):
    '''
    Subclass the original trainer for indiv_outcome.sequence models.
    '''
    def prepare_input(self, data):
        return self.model._prepare_input(data)

    def evaluate(self):
        mode = self.model.config['mode']
        metric_fn = METRICS_DICT[mode]
        res = self.model._predict_on_dataloader(self.test_dataloader)
        if res['pred'].shape != res['label']:
            res['pred'] = res['pred'].reshape(res['label'].shape)
        eval_res = metric_fn(res['pred'], res['label'])
        return eval_res

    def get_test_dataloader(self, test_data):
        return self.model.get_test_dataloader(test_data)



'''
input:
	data/raw_data.csv

output: 
	data/icdcode2ancestor_dict.pkl (icdcode to its ancestors)
	icdcode_embedding 

'''

import csv, re, pickle, os
import pdb
from functools import reduce 
import wget
from collections import defaultdict

import pandas as pd
import torch 
from torch import nn 
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
import icd10

ICDCODE_ANCESTOR_URL = 'https://github.com/futianfan/clinical-trial-outcome-prediction/raw/main/data/raw_data.csv'
icdcode2ancestor_dict_url = "https://storage.googleapis.com/pytrial/HINT-benchmark-data/icdcode2ancestor_dict.pkl"

def text_2_lst_of_lst(text):
	"""
		"[""['F53.0', 'P91.4', 'Z13.31', 'Z13.32']""]"
	"""
	text = text[2:-2]
	code_sublst = []
	for i in text.split('", "'):
		i = i[1:-1]
		code_sublst.append([j.strip()[1:-1] for j in i.split(',')])
	# print(code_sublst)	
	return code_sublst 

def get_icdcode_lst():
	input_file = 'demo_data/demo_trial_outcome_data/icdcode_raw_data.csv'
	if not os.path.exists(input_file):
		df = pd.read_csv(ICDCODE_ANCESTOR_URL)
		output_dir = os.path.dirname(input_file)
		if not os.path.exists(output_dir):
			os.makedirs(output_dir)
		df.to_csv(input_file, index=False)
	else:
		df = pd.read_csv(input_file)

	code_lst = []
	for _, row in df.iterrows():
		code_sublst = text_2_lst_of_lst(row['icdcodes'])
		code_lst.append(code_sublst)
	return code_lst 

def combine_lst_of_lst(lst_of_lst):
	lst = list(reduce(lambda x,y:x+y, lst_of_lst))
	lst = list(set(lst))
	return lst

def collect_all_icdcodes():
	code_lst = get_icdcode_lst()
	code_lst = list(map(combine_lst_of_lst, code_lst))
	code_lst = list(reduce(lambda x,y:x+y, code_lst))
	code_lst = list(set(code_lst))
	return code_lst


def find_ancestor_for_icdcode(icdcode, icdcode2ancestor):
	if icdcode in icdcode2ancestor:
		return 
	icdcode2ancestor[icdcode] = []
	ancestor = icdcode[:]
	while len(ancestor) > 2:
		ancestor = ancestor[:-1]
		if ancestor[-1]=='.':
			ancestor = ancestor[:-1]
		if icd10.find(ancestor) is not None:
			icdcode2ancestor[icdcode].append(ancestor)
	return


def build_icdcode2ancestor_dict():
	pkl_file = "pretrained_model/icdcode2ancestor/icdcode2ancestor_dict.pkl"

	if not os.path.exists('pretrained_model/icdcode2ancestor'):
		os.makedirs('pretrained_model/icdcode2ancestor')
		wget.download(icdcode2ancestor_dict_url, pkl_file)

	if os.path.exists(pkl_file):
		icdcode2ancestor = pickle.load(open(pkl_file, 'rb'))
		return icdcode2ancestor
	all_code = collect_all_icdcodes()
	icdcode2ancestor = defaultdict(list)
	for code in all_code:
		find_ancestor_for_icdcode(code, icdcode2ancestor)
	pickle.dump(icdcode2ancestor, open(pkl_file,'wb'))
	return icdcode2ancestor 


def collect_all_code_and_ancestor():
	icdcode2ancestor = build_icdcode2ancestor_dict()
	all_code = set(icdcode2ancestor.keys())
	ancestor_lst = list(icdcode2ancestor.values())
	ancestor_set = set(reduce(lambda x,y:x+y, ancestor_lst))
	all_code_lst = all_code.union(ancestor_set)
	return all_code_lst	


'''

assign each code an index. 

embedding lookup 


'''


class GRAM(nn.Sequential):
	"""	
		return a weighted embedding 
	"""

	def __init__(self, embedding_dim, icdcode2ancestor, device):
		super(GRAM, self).__init__()
		if icdcode2ancestor is None: 
			icdcode2ancestor = build_icdcode2ancestor_dict()

		self.icdcode2ancestor = icdcode2ancestor 
		self.all_code_lst = GRAM.codedict_2_allcode(self.icdcode2ancestor)
		self.code_num = len(self.all_code_lst)
		self.maxlength = 5
		self.code2index = {code:idx for idx,code in enumerate(self.all_code_lst)}
		self.index2code = {idx:code for idx,code in enumerate(self.all_code_lst)}
		self.padding_matrix = torch.zeros(self.code_num, self.maxlength).long() 
		self.mask_matrix = torch.zeros(self.code_num, self.maxlength)
		for idx in range(self.code_num):
			code = self.index2code[idx]
			ancestor_code_lst = self.icdcode2ancestor[code]
			ancestor_idx_lst = [idx] + [self.code2index[code] for code in ancestor_code_lst]
			ancestor_mask_lst = [1 for i in ancestor_idx_lst] + [0] * (self.maxlength - len(ancestor_idx_lst))
			ancestor_idx_lst = ancestor_idx_lst + [0]*(self.maxlength-len(ancestor_idx_lst))
			self.padding_matrix[idx,:] = torch.Tensor(ancestor_idx_lst)
			self.mask_matrix[idx,:] = torch.Tensor(ancestor_mask_lst)

		self.embedding_dim = embedding_dim 
		self.embedding = nn.Embedding(self.code_num, self.embedding_dim)
		self.attention_model = nn.Linear(2*embedding_dim, 1)

		self.device = device
		self = self.to(device)
		self.padding_matrix = self.padding_matrix.to('cpu')
		self.mask_matrix = self.mask_matrix.to('cpu')

	@property
	def embedding_size(self):
		return self.embedding_dim


	@staticmethod
	def codedict_2_allcode(icdcode2ancestor):
		all_code = set(icdcode2ancestor.keys())
		ancestor_lst = list(icdcode2ancestor.values())
		ancestor_set = set(reduce(lambda x,y:x+y, ancestor_lst))
		all_code_lst = all_code.union(ancestor_set)
		return all_code_lst		


	def forward_single_code(self, single_code):
		idx = self.code2index[single_code].to(self.device)
		ancestor_vec = self.padding_matrix[idx,:]  #### (5,)
		mask_vec = self.mask_matrix[idx,:] 

		embeded = self.embedding(ancestor_vec)  ### 5, 50
		current_vec = torch.cat([self.embedding(torch.Tensor([idx]).long()).view(1,-1) for i in range(self.maxlength)], 0) ### 1,50 -> 5,50
		attention_input = torch.cat([embeded, current_vec], 1)  ### 5, 100
		attention_weight = self.attention_model(attention_input)  ##### 5, 1
		attention_weight = torch.exp(attention_weight)  #### 5, 1
		attention_output = attention_weight * mask_vec.view(-1,1)  #### 5, 1
		attention_output = attention_output / torch.sum(attention_output)  #### 5, 1
		output = embeded * attention_output ### 5, 50 
		output = torch.sum(output, 0) ### 50
		return output 


	def forward_code_lst(self, code_lst):
		"""
			
			['C05.2', 'C10.0', 'C16.0', 'C16.4', 'C17.0', 'C17.1', 'C17.2'], length is 32 
			32 is length of code_lst; 5 is maxlength; 50 is embedding_dim; 
		"""
		idx_lst = [self.code2index[code] for code in code_lst if code in self.code2index] ### 32 
		if idx_lst == []:
			idx_lst = [0]
		ancestor_mat = self.padding_matrix[idx_lst,:].to(self.device)  ##### 32,5
		mask_mat = self.mask_matrix[idx_lst,:].to(self.device)  #### 32,5
		embeded = self.embedding(ancestor_mat)  #### 32,5,50
		current_vec = self.embedding(torch.Tensor(idx_lst).long().to(self.device)) #### 32,50
		current_vec = current_vec.unsqueeze(1) ### 32,1,50
		current_vec = current_vec.repeat(1, self.maxlength, 1) #### 32,5,50
		attention_input = torch.cat([embeded, current_vec], 2)  #### 32,5,100
		attention_weight = self.attention_model(attention_input)  #### 32,5,1 
		attention_weight = torch.exp(attention_weight).squeeze(-1)  #### 32,5 
		attention_output = attention_weight * mask_mat  #### 32,5 
		attention_output = attention_output / torch.sum(attention_output, 1).view(-1,1)  #### 32,5 
		attention_output = attention_output.unsqueeze(-1)  #### 32,5,1 
		output = embeded * attention_output ##### 32,5,50 
		output = torch.sum(output,1) ##### 32,50
		return output 

	def forward_code_lst2(self, code_lst_lst):
		### in one sample 
		code_lst = reduce(lambda x,y:x+y, code_lst_lst)
		code_embed = self.forward_code_lst(code_lst)
		### to do 
		code_embed = torch.mean(code_embed, 0).view(1,-1)  #### dim, 
		return code_embed 
		
	def forward_code_lst3(self, code_lst_lst_lst):
		code_embed_lst = [self.forward_code_lst2(code_lst_lst) for code_lst_lst in code_lst_lst_lst]
		code_embed = torch.cat(code_embed_lst, 0)
		return code_embed 





if __name__ == '__main__':
	dic = build_icdcode2ancestor_dict() 




# if __name__ == "__main__":
# 	# code_lst = collect_all_icdcodes()  ### 5k code
# 	# all_code = collect_all_code_and_ancestor() ### 10k 
# 	# icdcode2ancestor = build_icdcode2ancestor_dict()
# 	# maxlength = 0
# 	# for icdcode, ancestor in icdcode2ancestor.items():
# 	# 	if len(ancestor) > maxlength:
# 	# 		maxlength = len(ancestor)
# 	# print(maxlength) 
# 	# assert maxlength == 4

# 	icdcode2ancestor = build_icdcode2ancestor_dict()
# 	gram_model = GRAM(embedding_dim = 50, icdcode2ancestor = icdcode2ancestor)
# 	# output = gram_model.single_forward('S33.121S')
# 	code_lst = ['C05.2', 'C10.0', 'C16.0', 'C16.4', 'C17.0', 'C17.1', 'C17.2']
# 	output = gram_model(code_lst)










import pdb
import math
from collections import defaultdict

import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence, pack_padded_sequence, pad_packed_sequence
import numpy as np
from tqdm import trange

from pytrial.data.patient_data import SequencePatientBase, SeqPatientCollator
from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import SequenceSimulationBase
from ..losses import GeneratorLoss, DiscriminatorLossGP, MultilabelBinaryXentLoss
from ..trainer import SeqSimSyntegTrainer
from ..data import SequencePatient

class Embedding(nn.Module):
    def __init__(self, total_vocab_size, emb_size):
        """Construct an embedding matrix to embed sparse codes"""
        super(Embedding, self).__init__()
        self.code_embed = nn.Embedding(total_vocab_size+3, emb_size)

    def forward(self, codes): # batch_size * visits * codes
        code_embeds = self.code_embed(codes)
        return code_embeds

class SingleVisitTransformer(nn.Module):
    """An Encoder Transformer to turn code embeddings into a visit embedding"""
    def __init__(self, emb_size, n_head, hidden_dim):
        super(SingleVisitTransformer, self).__init__()
        encoderLayer = nn.TransformerEncoderLayer(emb_size, n_head, 
                        dim_feedforward=hidden_dim, dropout=0.1, activation="relu", 
                        layer_norm_eps=1e-08, batch_first=True)
        self.transformer = nn.TransformerEncoder(encoderLayer, 2)

    def forward(self, code_embeddings, visit_lengths):
        bs, vs, cs, ed = code_embeddings.shape
        mask = torch.ones((bs, vs, cs)).to(code_embeddings.device)
        for i in range(bs):
            for j in range(vs):
                mask[i,j,:visit_lengths[i,j]] = 0
        visits = torch.reshape(code_embeddings, (bs*vs,cs,ed))
        mask = torch.reshape(mask, (bs*vs,cs))
        encodings = self.transformer(visits, src_key_padding_mask=mask)
        encodings = torch.reshape(encodings, (bs,vs,cs,ed))
        visit_representations = encodings[:,:,0,:]
        return visit_representations

class RecurrentLayer(nn.Module):
    """An Recurrent Layer to predict the next visit based on the visit embeddings"""
    def __init__(self, hidden_dim, n_rnn_layer):
        super(RecurrentLayer, self).__init__()
        self.lstm = nn.LSTM(input_size=hidden_dim, hidden_size=hidden_dim, num_layers=n_rnn_layer, dropout=0.1)

    def forward(self, visit_embeddings):   
        output, _ = self.lstm(visit_embeddings)
        return output

class DependencyModel(nn.Module):
    """The entire Dependency Model component of SynTEG"""
    def __init__(self, emb_size, hidden_dim, condition_dim, total_vocab_size, n_head, n_rnn_layer):
        super(DependencyModel, self).__init__()
        self.embeddings = Embedding(total_vocab_size, emb_size)
        self.visit_att = SingleVisitTransformer(emb_size, n_head, hidden_dim)
        self.proj1 = nn.Linear(emb_size, hidden_dim)
        self.lstm = RecurrentLayer(hidden_dim, n_rnn_layer)
        self.proj2 = nn.Linear(hidden_dim, condition_dim)
        self.proj3 = nn.Linear(condition_dim, total_vocab_size)
        
    def forward(self, inputs_word, visit_lengths, export=False):  # bs * visits * codes, bs * visits * 1 
        inputs = self.embeddings(inputs_word) # bs * visits * codes * emb_size
        inputs = self.visit_att(inputs, visit_lengths) # bs * visits * emb_size
        inputs = self.proj1(inputs) # bs * visits * hidden_dim
        output = self.lstm(inputs) # bs * visits * hidden_dim
        if export:
            return self.proj2(output)[:, :-1, :] # bs * visit * condition
        else:
            output = self.proj3(torch.relu(self.proj2(output))) # bs * visits * total_vocab_size
            diagnosis_output = output[:, :-1, :]
            return diagnosis_output

#######
### Conditional GAN Model
#######

class PointWiseLayer(nn.Module):
    def __init__(self, num_outputs):
        """Construct an embedding matrix to embed sparse codes"""
        super(PointWiseLayer, self).__init__()
        self.bias = nn.Parameter(torch.zeros(num_outputs).uniform_(-math.sqrt(num_outputs), math.sqrt(num_outputs)))

    def forward(self, x1, x2):
        return x1 * x2 + self.bias

class Generator(nn.Module):
    def __init__(self, g_dims, z_dim, condition_dim):
        super(Generator, self).__init__()
        self.dense_layers = nn.Sequential(*[nn.Linear(g_dims[i-1] if i > 0 else z_dim, g_dims[i]) for i in range(len(g_dims[:-1]))])
        self.batch_norm_layers = nn.Sequential(*[nn.BatchNorm1d(dim, eps=1e-5) for dim in g_dims[:-1]])
        self.output_layer = nn.Linear(g_dims[-2], g_dims[-1])
        self.output_sigmoid = nn.Sigmoid()
        self.condition_layers = nn.Sequential(*[nn.Linear(condition_dim, dim) for dim in g_dims[:-1]])
        self.pointwiselayers = nn.Sequential(*[PointWiseLayer(dim) for dim in g_dims[:-1]])

    def forward(self, x, condition):
        for i in range(len(self.dense_layers)):
            h = self.dense_layers[i](x)
            x = nn.functional.relu(self.pointwiselayers[i](self.batch_norm_layers[i](h), self.condition_layers[i](condition)))
        x = self.output_layer(x)
        x = self.output_sigmoid(x)
        return x

class Discriminator(nn.Module):
    def __init__(self, d_dims, g_dims, condition_dim):
        super(Discriminator, self).__init__()
        self.dense_layers = nn.Sequential(*[nn.Linear(d_dims[i-1] if i > 0 else g_dims[-1] + 1, d_dims[i]) for i in range(len(d_dims))])
        self.layer_norm_layers = nn.Sequential(*[nn.LayerNorm(dim, eps=1e-5) for dim in d_dims])
        self.output_layer = nn.Linear(d_dims[-1], 1)
        self.condition_layers = nn.Sequential(*[nn.Linear(condition_dim, dim) for dim in d_dims])
        self.pointwiselayers = nn.Sequential(*[PointWiseLayer(dim) for dim in d_dims])

    def forward(self, x, condition):
        a = (2 * x) ** 15
        sparsity = torch.sum(a / (a + 1), axis=-1, keepdim=True)
        x = torch.cat((x, sparsity), axis=-1)
        for i in range(len(self.dense_layers)):
            h = self.dense_layers[i](x)
            x = self.pointwiselayers[i](self.layer_norm_layers[i](h), self.condition_layers[i](condition))
        x = self.output_layer(x)
        return x
      
class BuildModel(nn.Module):
    def __init__(self,
        emb_size,
        hidden_dim,
        condition_dim,
        n_head,
        n_rnn_layer,
        total_vocab_size,
        z_dim,
        g_dims,
        d_dims,
        device,
        max_visit,
        max_code_per_visit,
        **kwargs,
        ) -> None:
        super().__init__()
        
        self.device = device
        self.total_vocab_size = total_vocab_size
        self.max_visit = max_visit
        self.max_code_per_visit = max_code_per_visit
        self.z_dim = z_dim
        
        self.dependency_module = DependencyModel(
            emb_size=emb_size,
            hidden_dim=hidden_dim,
            condition_dim=condition_dim,
            total_vocab_size=total_vocab_size,
            n_head=n_head,
            n_rnn_layer=n_rnn_layer
        )

        self.generator_module = Generator(
            z_dim=z_dim,
            g_dims=g_dims,
            condition_dim=condition_dim
        )
        
        self.discriminator_module = Discriminator(
            d_dims=d_dims,
            g_dims=g_dims,
            condition_dim=condition_dim
        )
        
    def dependency_forward(self, inputs, export=False):
        logits = self.dependency_module(inputs['v'], inputs['c_lengths'], export)
        if export:
            return logits
        
        labels = torch.sum(nn.functional.one_hot(inputs['v'].long(), num_classes=self.total_vocab_size+2), dim=2).float()[:, 1:, :-2]
        for idx in range(len(inputs['v_lengths'])):
            logits[idx, inputs['v_lengths'][idx]:] = 0
        
        return logits, labels
    
    def gan_forward(self, inputs):
        x_real = torch.sum(nn.functional.one_hot(inputs['v'].long(), num_classes=self.total_vocab_size+2), dim=1).float()[:, :-2]
        z = torch.randn((len(x_real), self.z_dim)).to(self.device)
        x_fake = self.generator_module(z, inputs['conditions'])
        y_fake = self.discriminator_module(x_fake, inputs['conditions'])
        y_real = self.discriminator_module(x_real, inputs['conditions'])
        alpha = torch.rand((x_real.size(0), 1)).to(self.device)
        interpolates = (alpha * x_real + (1-alpha)*x_fake).requires_grad_(True).float()
        d_interpolates = self.discriminator_module(interpolates, inputs['conditions'])
        return {
            'y_fake': y_fake,
            'y_real': y_real,
            'interpolates': interpolates,
            'd_interpolates': d_interpolates
        }
    
    def forward(self, inputs, export=False):
        if 'conditions' in inputs:
            return self.gan_forward(inputs)
        else:
            return self.dependency_forward(inputs, export)
        
    def add_condition(self, inputs):
        condition = self.dependency_forward(inputs, True)
        visits = inputs['v'][:,1:]
        v_new = []
        condition_new = []
        for i in range(len(inputs['v_lengths'])):
            for j in range(inputs['v_lengths'][i]):
                v_new.append(visits[i, j, :])
                condition_new.append(condition[i,j,:])
        
        inputs['v'] = torch.stack(v_new)
        inputs['conditions'] = torch.stack(condition_new)
        return inputs
      
    def sample(self, n_samples):
        ehr = torch.zeros((n_samples, 1, self.total_vocab_size), device=self.device)
        batch_ehr = torch.ones((n_samples, self.max_visit+1, self.max_code_per_visit+1), device=self.device, dtype=torch.int) * (self.total_vocab_size + 1)
        batch_ehr[:,:,0] = self.total_vocab_size
        batch_lens = torch.zeros((n_samples, self.max_visit+1, 1), dtype=torch.int, device=self.device)
        batch_lens[:,0,0] = 1
        with torch.no_grad():
            for j in trange(self.max_visit):
                for i in range(n_samples):
                    codes = torch.nonzero(ehr[i,j]).squeeze(1)
                    batch_ehr[i,j,:min(len(codes), self.max_code_per_visit)] = codes[1:min(len(codes), self.max_code_per_visit) + 1]
                    batch_lens[i,j] = 1 + min(len(codes), self.max_code_per_visit)
                
                condition_vector = self.dependency_module(batch_ehr, batch_lens, export=True)
                condition = condition_vector[:,j,:]
                z = torch.randn((n_samples, self.z_dim), device=self.device)
                visit = self.generator_module(z, condition)
                visit = torch.bernoulli(visit).unsqueeze(1)
                ehr = torch.cat((ehr, visit), dim=1)
        return ehr[:,1:]
    

class SynTEG(SequenceSimulationBase):
    '''
    Implement a GAN based model for longitudinal patient records simulation [1]_.
    
    Parameters
    ----------
    vocab_size: list[int]
        A list of vocabulary size for different types of events, e.g., for diagnosis, procedure, medication.

    order: list[str]
        The order of event types in each visits, e.g., ``['diag', 'prod', 'med']``.
        Visit = [diag_events, prod_events, med_events], each event is a list of codes.

    max_visit: int
        Maximum number of visits.
        
    max_code_per_visit: int
        Maximum number of medical codes in a single visit.

    emb_size: int
        Embedding size for encoding input event codes.
        
    hidden_dim: int
        Size of intermediate hidden dimension for RNN and Feed Forward layers
        
    condition_dim: int
        Size of intermediate dimension for encoding medical history to condition the GAN
        
    n_head: int
        Number of attention heads
        
    n_rnn_layer: int
        Number of RNN layers for encoding historical events.
        
    z_dim: int
        Dimension of noise vector passed into the GAN Generator
        
    g_dims: list
        List of ints for intermediate GAN Generator dimensionalities
        
    d_dims: list
        List of ints for intermediate GAN Discriminator dimensionalities

    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    Notes
    -----
    .. [1] Zhang, Ziqi, et al. (2021, March). SynTEG: a framework for temporal structured electronic health data simulation. Journal of the American Medical Informatics Association 28.3.
    '''
    def __init__(self,
        vocab_size,
        order,
        max_visit=20,
        max_code_per_visit=20,
        emb_size=64,
        hidden_dim=32,
        condition_dim=32,
        n_head=4,
        n_rnn_layer=2,
        z_dim=64,
        g_dims=[32, 32, 64, 64],
        d_dims=[64, 32, 32],
        learning_rate=1e-3,
        batch_size=64,
        epochs=20,
        num_worker=0,
        device='cpu',# 'cuda:0',
        experiment_id='trial_simulation.sequence.synteg',
        ):
        super().__init__(experiment_id)
        self.config = {
            'vocab_size': vocab_size, 
            'orders': order, 
            'max_visit': max_visit,
            'max_code_per_visit': max_code_per_visit,
            'emb_size': emb_size,
            'hidden_dim': hidden_dim,
            'condition_dim': condition_dim,
            'n_head': n_head,
            'n_rnn_layer': n_rnn_layer,
            'z_dim': z_dim,
            'g_dims': g_dims,
            'd_dims': d_dims,
            'learning_rate': learning_rate,
            'batch_size': batch_size,
            'epochs': epochs,
            'num_worker': num_worker,
            'device':device,
            }
        self.config['total_vocab_size'] = sum(vocab_size)
        self.config['g_dims'].append(self.config['total_vocab_size'])
        self.device = device
        self._build_model()
        
    def _build_model(self):
        self.model = BuildModel(
            emb_size = self.config['emb_size'],
            hidden_dim = self.config['hidden_dim'],
            condition_dim = self.config['condition_dim'],
            n_head = self.config['n_head'],
            n_rnn_layer = self.config['n_rnn_layer'],
            total_vocab_size = self.config['total_vocab_size'],
            z_dim = self.config['z_dim'],
            g_dims = self.config['g_dims'],
            d_dims = self.config['d_dims'],
            device = self.device,
            max_visit = self.config['max_visit'],
            max_code_per_visit = self.config['max_code_per_visit']
            )
        self.model = self.model.to(self.device)
        
    def fit(self, train_data):
        '''
        Train model with sequential patient records.

        Parameters
        ----------
        train_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events.
        '''
        self._input_data_check(train_data)
        self._fit_model(train_data)

    def predict(self, n, return_tensor=True):
        '''
        Generate synthetic records

        Parameters
        ----------
        n: int
            How many samples in total will be generated.
        
        return_tensor: bool
            If `True`, return output generated records in tensor format (n, n_visit, n_event), good for later predictive modeling.
            If `False, return records in `SequencePatient` format.
        '''
        assert isinstance(n, int), 'Input `n` should be integer.'
        outputs = self._predict(n)
        if not return_tensor:
            outputs = self._translate_sparse_visits_to_dense(outputs)
        return outputs

    def save_model(self, output_dir):
        '''
        Save the learned simulation model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
            
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({
                  'dependency': self.model.dependency_module.state_dict(),
                  'generator': self.model.generator_module.state_dict(),
                  'discriminator': self.model.discriminator_module.state_dict()
              }, output_dir=output_dir)

    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
            - If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            - If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename, map_location='cpu')
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        self.model.dependency_module.load_state_dict(state_dict['dependency'])
        self.model.generator_module.load_state_dict(state_dict['generator'])
        self.model.discriminator_module.load_state_dict(state_dict['discriminator'])
    
    def get_train_dataloader(self, train_data):
        dataloader = DataLoader(train_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=True,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':train_data.metadata['visit']['mode'],
                    'label_mode':train_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader
    
    def get_test_dataloader(self, test_data):
        dataloader = DataLoader(test_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=False,
            shuffle=False,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':test_data.metadata['visit']['mode'],
                    'label_mode':test_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader
    
    def _build_simulation_loss_model(self):
        return [MultilabelBinaryXentLoss(self.model)]
    
    def _build_gan_loss_model(self):
        return [DiscriminatorLossGP(self.model), GeneratorLoss(self.model)]

    def _fit_model(self, train_data):
        # PHASE 1: Simulation
        train_dataloader = self.get_train_dataloader(train_data)
        simulation_loss_models = self._build_simulation_loss_model()
        simulation_train_objectives = [(train_dataloader, loss_model) for loss_model in simulation_loss_models]
        trainer = SeqSimSyntegTrainer(
            model=self,
            train_objectives=simulation_train_objectives
            )
        trainer.train(**self.config)

        # PHASE 2: GAN
        gan_loss_models = self._build_gan_loss_model()
        gan_train_objectives = [(train_dataloader, loss_model) for loss_model in gan_loss_models]
        trainer = SeqSimSyntegTrainer(
            model=self,
            train_objectives=gan_train_objectives,
            condition=True
            )
        trainer.train(**self.config)

    @torch.no_grad()
    def _predict(self, n):
        return self.model.sample(n)
    
    def _translate_sparse_visits_to_dense(self, visits):
        def _map_func(x):
            res = np.where(x > 0)[0].tolist()
            return [0] if len(res) == 0 else res # pad if nothing happened

        outputs = defaultdict(list)
        for batchv in visits:
            voc_offset = 0
            for i, o in enumerate(self.config['orders']):
                voc_size = self.config['vocab_size'][i]
                visit = batchv[...,voc_offset:voc_offset+voc_size]
                visit = visit.cpu().numpy()
                res = list(map(_map_func, visit))
                outputs[o].append(res)
                voc_offset += voc_size

        n_total = len(outputs[o])
        sample_list = []        
        for i in range(n_total):
            sample = []
            for numv in range(len(outputs[o][i])):
                visit = []
                for o in self.config['orders']:
                    visit.append(outputs[o][i][numv])
                sample.append(visit)
            sample_list.append(sample)

        # create seqpatient data
        return SequencePatient(
            data={'v':sample_list},
            metadata={
                'visit':{'mode':'dense','order':self.config['orders']},
                }
            )

    def _translate_dense_visits_to_sparse(self, visits):
        total_vocab_size = sum(self.config['vocab_size'])
        num_visits = len(visits[self.config['orders'][0]])
        outputs = np.zeros((num_visits, total_vocab_size))

        for i, o in enumerate(self.config['orders']):
            for j in range(num_visits):
                raw = visits[o][j]
                if isinstance(raw, torch.Tensor): raw = raw.detach().cpu().numpy()
                if i > 0: 
                    voc_size = sum(self.config['vocab_size'][:i-1])
                    if isinstance(raw, list):
                        raw = [r + voc_size for r in raw]
                    else:
                        raw += voc_size
                outputs[j, raw] = 1

        return outputs

    def _pad_multiple_tensor_visits(self, visits):
        new_list = []
        for v in visits:
            new_list.extend([torch.tensor(x).squeeze(0) for x in np.array_split(v, len(v))])
        return pad_sequence(new_list, batch_first=True)
    
    def _prepare_input(self, data):
        '''
        Prepare inputs for sequence simulation models.

        Parameters
        ----------
        data: dict[list]
            A batch of patient records.
        '''
        visits = data['v']
        feature = data['x']
        if not isinstance(feature, torch.Tensor): feature = torch.tensor(feature)

        feature = feature.to(self.device)

        inputs = {
            'v':{},
            'c_lengths':{},
            'v_lengths':[],
            'x':feature, # baseline feature
            }
        
        v_lengths = [len(visits[self.config['orders'][0]][idx][:self.config['max_visit']]) for idx in range(len(visits[self.config['orders'][0]]))]
        inputs['v_lengths'] = v_lengths

        v = torch.ones(len(v_lengths), max(v_lengths) + 1, self.config['max_code_per_visit'] + 1, dtype=torch.int) * (self.config['total_vocab_size'] + 1)
        c_lengths = torch.ones(len(v_lengths), max(v_lengths) + 1, 1, dtype=torch.int)
        for i in range(len(v_lengths)):
            for j in range(min(v_lengths[i], self.config['max_visit'])):
                visit = torch.IntTensor([r + sum(self.config['vocab_size'][:n-1]) if n > 0 else r for n, o in enumerate(self.config['orders']) for r in visits[o][i][j]][:self.config['max_code_per_visit']])
                v[i, j, 1:len(visit)+1] = visit
                c_lengths[i, j] = len(visit) + 1
        v[:,:,0] = self.config['total_vocab_size']
        c_lengths[:,0] = 1
        v = v.to(self.device)
        c_lengths = c_lengths.to(self.device)
        inputs['v'] = v
        inputs['c_lengths'] = c_lengths
        
        return inputs
    
    def _add_condition(self, inputs):
        return self.model.add_condition(inputs)
    
    def _input_data_check(self, inputs):
        assert isinstance(inputs, SequencePatientBase), f'`trial_simulation.sequence` models require input training data in `SequencePatientBase`, find {type(inputs)} instead.'
        


'''
Tools to download data and process data from clinicaltrials.gov.
Part of codes come from pytrials: https://pytrials.readthedocs.io/en/latest/index.html.
'''
import pdb
import requests
import csv
import re
import os
import pdb
import time
import zipfile

import pandas as pd
import wget

class ClinicalTrials:
    '''Utilities for download and preprocess trial datasets from ClinicalTrial.gov.
    '''
    __raw_txt_dir__ = './aact-raw'
    _BASE_URL = "https://clinicaltrials.gov/api/"
    _INFO = "info/"
    _QUERY = "query/"
    _JSON = "fmt=json"
    _CSV = "fmt=csv"

    def __init__(self) -> None:
        self.api_info = self._api_info()

    @property
    def study_fields(self):
        '''Get the all possible fields to get from ClinicalTrials.gov.
        '''
        fields_list = json_handler(
            f"{self._BASE_URL}{self._INFO}study_fields_list?{self._JSON}"
        )
        return fields_list["StudyFields"]["Fields"]

    def load_data(self,
        input_dir='./datasets/AACT-ClinicalTrial'
        ):
        '''Load the pre-downloaded and processed csv trial documents.

        Parameters
        ----------

        input_dir: str
            The dir of the input data, should have a csv under named 'clinical_trials.csv';
            Or the path direct to the csv file.
        '''
        if os.path.isdir(input_dir):
            input_dir = os.path.join(input_dir, 'clinical_trials.csv')
        df = pd.read_csv(input_dir, index_col=0)
        return df

    def download(self,
        date='20220501',
        output_dir='./datasets/AACT-ClinicalTrial',
        ):
        '''Download a static copy of all clinical trial documents from clinicaltrials.gov.

        Parameters
        ----------
        date: str
            The date of the database copy.

        fields: list[str]
            A list of fields should be included in the downloaded dataframe.

        output_dir: str
            The output directory of the downloaded data.
        '''
        url = f'https://aact.ctti-clinicaltrials.org/static/exported_files/daily/{date}_pipe-delimited-export.zip'
        if not os.path.exists(output_dir): os.makedirs(output_dir)
        print(f'Download all the clinical trial records from {url}, save to {output_dir}.')
        filename = wget.download(url, out=os.path.join(output_dir,'./aact-raw.zip'))
        zipf = zipfile.ZipFile(filename, 'r')
        zipf.extractall(os.path.join(output_dir, self.__raw_txt_dir__))
        zipf.close()
        self.process_studies(input_dir=os.path.join(output_dir, self.__raw_txt_dir__),
            output_dir=output_dir)

    def query_studies(self,
        search_expr,
        fields=['NCTId', 'Condition', 'InterventionName', 'Keyword', 'PrimaryOutcomeMeasure', 'BriefTitle', 'EligibilityCriteria', 'DetailedDescription', 'OverallStatus'],
        max_studies=50,
        fmt='csv'):
        '''Query study content for specified fields from the remote clinicaltrial.gov API.
        Retrieves information from the study fields endpoint, which acquires specified information
        from a large (max 1000) studies. To see a list of all possible fields, check the class'
        study_fields attribute.

        Parameters
        ----------
            search_expr (str): A string containing a search expression as specified by
                `their documentation <https://clinicaltrials.gov/api/gui/ref/syntax#searchExpr>`_.

            fields (list(str)): A list containing the desired information fields.

            max_studies (int): An integer indicating the maximum number of studies to return.
                Defaults to 50.

            fmt (str): A string indicating the output format, csv or json. Defaults to csv.

        Returns
        -------
            Either a dict, if fmt='json', or a list of records (e.g. a list of lists), if fmt='csv.
            Both containing the maximum number of study fields queried using the specified search expression.

        Raises
        ------
            ValueError: The number of studies can only be between 1 and 1000
            ValueError: One of the fields is not valid! Check the study_fields attribute
                for a list of valid ones.
            ValueError: Format argument has to be either 'csv' or 'json'
        '''

        if max_studies > 1000 or max_studies < 1:
            raise ValueError("The number of studies can only be between 1 and 1000")
        elif not set(fields).issubset(self.study_fields):
            raise ValueError(
                "One of the fields is not valid! Check the study_fields attribute for a list of valid ones."
            )
        else:
            concat_fields = ",".join(fields)
            req = f"study_fields?expr={search_expr}&max_rnk={max_studies}&fields={concat_fields}"

            if fmt == "csv":
                url = f"{self._BASE_URL}{self._QUERY}{req}&{self._CSV}"
                fields = csv_handler(url)
                return pd.DataFrame.from_records(fields[1:], columns=fields[0])

            elif fmt == "json":
                url = f"{self._BASE_URL}{self._QUERY}{req}&{self._JSON}"
                return json_handler(url)

            else:
                raise ValueError("Format argument has to be either 'csv' or 'json'")

    def get_study_count(self, search_expr):
        """Returns study count for specified search expression
        Retrieves the count of studies matching the text entered in search_expr.

        Parameters
        ----------
            search_expr (str): A string containing a search expression as specified by
                `their documentation <https://clinicaltrials.gov/api/gui/ref/syntax#searchExpr>`_.

        Returns
        -------
            An integer

        Raises
        ------
            ValueError: The search expression cannot be blank.
        """
        if not set(search_expr):
            raise ValueError("The search expression cannot be blank.")
        else:
            req = f"study_fields?expr={search_expr}&max_rnk=1&fields=NCTId"
            url = f"{self._BASE_URL}{self._QUERY}{req}&{self._JSON}"
            returned_data = json_handler(url)
            study_count = returned_data["StudyFieldsResponse"]["NStudiesFound"]
            return study_count

    def _api_info(self):
        last_updated = json_handler(
            f"{self._BASE_URL}{self._INFO}data_vrs?{self._JSON}"
        )["DataVrs"]
        api_version = json_handler(f"{self._BASE_URL}{self._INFO}api_vrs?{self._JSON}")[
            "APIVrs"
        ]
        return api_version, last_updated

    def process_studies(self,
        input_dir='./datasets/aact-raw',
        output_dir='./datasets'
        ):
        '''Process the raw separate delimited trial documents and combine to a complete csv file.
        '''
        start_time = time.time()
        print('processing description {:.1f} sec'.format(time.time() - start_time))
        df_merge = pd.read_csv(os.path.join(input_dir, 'brief_summaries.txt'), sep='|')
        df_merge = df_merge[['nct_id', 'description']]

        print('processing studies {:.1f} sec'.format(time.time() - start_time))
        df = pd.read_csv(os.path.join(input_dir, 'studies.txt'), sep='|')
        df = df[['nct_id', 'brief_title']]
        df = df.groupby('nct_id').apply(lambda x: ', '.join(map(str,x['brief_title'])))
        df = df.reset_index(drop=False)
        df = df.rename(columns={0:'title'})
        df_merge = pd.merge(df_merge, df, on='nct_id', how='outer')
        df_merge['title'].fillna(value='none',inplace=True)

        print('processing interventions {:.1f} sec'.format(time.time() - start_time))
        df = pd.read_csv(os.path.join(input_dir, 'interventions.txt'), sep='|')
        df = df[['nct_id','name']]
        df = df.groupby('nct_id').apply(lambda x: ', '.join(map(str,x['name'])))
        df = df.reset_index(drop=False)
        df = df.rename(columns={0:'intervention_name'})
        df_merge = pd.merge(df_merge, df, on='nct_id', how='outer')
        df_merge['intervention_name'].fillna(value='none',inplace=True)

        print('processing conditions/diseases {:.1f} sec'.format(time.time() - start_time))
        df = pd.read_csv(os.path.join(input_dir, 'conditions.txt'), sep='|')
        df = df[['nct_id','name']]
        df = df.groupby('nct_id').apply(lambda x: ', '.join(map(str,x['name'])))
        df = df.reset_index(drop=False)
        df = df.rename(columns={0:'disease'})
        df_merge = pd.merge(df_merge, df, on='nct_id', how='outer')
        df_merge['disease'].fillna('none', inplace=True)

        print('processing keywords {:.1f} sec'.format(time.time() - start_time))
        df = pd.read_csv(os.path.join(input_dir, 'keywords.txt'), sep='|')
        df = df[['nct_id','name']]
        df = df.groupby('nct_id').apply(lambda x: ', '.join(map(str,x['name'])))
        df = df.reset_index(drop=False)
        df = df.rename(columns={0:'keyword'})
        df_merge = pd.merge(df_merge, df, on='nct_id', how='outer')
        df_merge['keyword'].fillna('none', inplace=True)

        print('processing outcomes {:.1f} sec'.format(time.time() - start_time))
        df = pd.read_csv(os.path.join(input_dir, 'design_outcomes.txt'), sep='|')
        df = df[['nct_id','measure',]]
        df = df.groupby('nct_id').apply(lambda x: ', '.join(map(str,x['measure'])))
        df = df.reset_index(drop=False)
        df = df.rename(columns={0:'outcome_measure'})
        df_merge = pd.merge(df_merge, df, on='nct_id', how='outer')
        df_merge['outcome_measure'].fillna('none', inplace=True)

        print('processing eligbility criteria {:.1f} sec'.format(time.time() - start_time))
        df = pd.read_csv(os.path.join(input_dir, 'eligibilities.txt'), sep='|')
        df = df[['nct_id','criteria']]
        df = df.groupby('nct_id').apply(lambda x: ', '.join(map(str,x['criteria'])))
        df = df.reset_index(drop=False)
        df = df.rename(columns={0:'criteria'})
        df_merge = pd.merge(df_merge, df, on='nct_id', how='outer')
        df_merge['criteria'].fillna('none', inplace=True)

        print('processing references {:.1f} sec'.format(time.time() - start_time))
        df = pd.read_csv(os.path.join(input_dir, 'study_references.txt'), sep='|')
        df = df[['nct_id','citation',]]
        df['citation'] = df['citation'].apply(lambda x: x.split('.')).apply(lambda x: x[1] if len(x)>1 else x[0])
        df = df.groupby('nct_id').apply(lambda x: ', '.join(map(str,x['citation'])))
        df = df.reset_index(drop=False)
        df = df.rename(columns={0:'reference'})
        df_merge = pd.merge(df_merge, df, on='nct_id', how='outer')
        df_merge['reference'].fillna('none', inplace=True)
        df_merge.loc[df_merge['reference'] == '']['reference'] = 'none'

        print('processing study status {:.1f} sec'.format(time.time() - start_time))
        df = pd.read_csv(os.path.join(input_dir,'studies.txt'), sep='|')
        df_merge = pd.merge(df_merge, df[['nct_id', 'overall_status']], on='nct_id', how='outer')
        df_merge['overall_status'].fillna('none', inplace=True)
        df_merge.fillna('none', inplace=True)
        if os.path.isdir(output_dir):
            save_path = os.path.join(output_dir, 'clinical_trials.csv')
        else:
            save_path = output_dir
        df_merge = df_merge[df_merge['overall_status'] != 'Withheld'].reset_index(drop=True)
        df_merge.to_csv(save_path)
        print(f'saving processed the csv file to {save_path}')
        return df_merge

    def __repr__(self):
        return f"ClinicalTrials.gov client v{self.api_info[0]}, database last updated {self.api_info[1]}"

def request_ct(url):
    """Performs a get request that provides a (somewhat) useful error message."""
    try:
        response = requests.get(url)
    except ImportError:
        raise ImportError(
            "Couldn't retrieve the data, check your search expression or try again later."
        )
    else:
        return response

def json_handler(url):
    """Returns request in JSON (dict) format"""
    return request_ct(url).json()

def csv_handler(url):
    """Returns request in CSV (list of records) format"""
    response = request_ct(url)
    decoded_content = response.content.decode("utf-8")
    split_by_blank = re.split(r"\n\s*\n", decoded_content)  # Extracts header info
    cr = csv.reader(split_by_blank[1].splitlines(), delimiter=",")
    records = list(cr)
    return records

def test():
    client = ClinicalTrials()
    print(client.study_fields)

    df = client.query_studies(search_expr='Coronavirus+COVID',
        max_studies=500,
        )

    # df = client.process_studies('./datasets/AACT-ClinicalTrial/aact-raw', './datasets/AACT-ClinicalTrial')
    # print(df.head())

if __name__ == '__main__':
    test()



'''
Implement Logistic Regression model for clinical trial outcome prediction
'''
import os
import json

from sklearn.metrics import roc_auc_score
import numpy as np 
import torch 
from torch import nn 
from torch.autograd import Variable
from rdkit import Chem 
from rdkit.Chem import AllChem
from rdkit.Chem import DataStructs

from .base import TrialOutcomeBase
from .model_utils.utils import trial_collate_fn

class BuildModel(nn.Module):
    def __init__(self,
        input_dim=512,
        output_dim=1,
        hidden_dim=128,
        num_layer=2,
        ) -> None:
        super().__init__()
        if num_layer == 1:
            self.mlp = nn.ModuleList([nn.Linear(input_dim, output_dim)])
        else:
            self.mlp = nn.ModuleList([nn.Linear(input_dim, hidden_dim)])

        for _ in range(num_layer-2):
            self.mlp.append(nn.ReLU())
            self.mlp.append(nn.Linear(hidden_dim, hidden_dim))
        
        if num_layer > 1:
            self.mlp.append(nn.ReLU())
            self.mlp.append(nn.Linear(hidden_dim, output_dim))
    
    def forward(self, inputs):
        h = inputs
        h = h.float()
        for layer in self.mlp:
            h = layer(h)
        return h

class MLP(TrialOutcomeBase):
    '''
    Implement MLP model for clinical trial outcome prediction. 

    Parameters
    ----------
    epoch: int
        number of training epochs.
    
    lr: float
        learning rate.

    weight_decay: float
        Regularization strength for l2 norm; must be a positive float.
    '''
    def __init__(self, 
        epoch = 5,
        lr = 1e-3,
        batch_size = 32, 
        weight_decay=0.0
        ):

        super(MLP, self).__init__()
        
        self.model = BuildModel() 

        self.config = {
            'epoch': epoch,
            'lr': lr,
            'weight_decay': weight_decay,
            'batch_size': batch_size,
        }

        self.epoch = epoch 
        self.lr = lr 
        self.weight_decay = weight_decay
        self.batch_size = batch_size
        self.loss = nn.BCEWithLogitsLoss()

    def feature(self, data_loader):
        labels = []
        features = []
        nctids = []
        for nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst in data_loader:
            label_vec = label_vec.to(self.device)
            nctids.extend(nctid_lst)
            for smiles_lst in smiles_lst2: 
                mol_lst = [Chem.MolFromSmiles(smiles) for smiles in smiles_lst]
                mol_lst = list(filter(lambda x:x is not None, mol_lst))
                fp_lst = [AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=512) for mol in mol_lst]
                feature = np.zeros(512)
                for fp in fp_lst:
                    arr = np.zeros((0,), dtype=np.int8)
                    DataStructs.ConvertToNumpyArray(fp,arr)
                    feature += arr 
                feature = feature.reshape(1,-1)
                features.append(feature)
        features = np.concatenate(features, 0)
        labels = np.array(labels)	
        return nctids, features, labels 	


    def fit(self, train_data, valid_data=None):
        '''
        Train model to predict clinical trial outcomes.

        Parameters
        ----------
        train_data: TrialOutcomeDatasetBase
            Training data, should be a `TrialOutcomeDatasetBase` object.
        
        valid_data: TrialOutcomeDatasetBase
            Validation data, should be a `TrialOutcomeDatasetBase` object. 
        '''

        # build dataloader using train_data
        train_loader = self._build_dataloader_from_dataset(train_data, 
            num_workers=0, batch_size=self.batch_size, shuffle=True,
            collate_fn=trial_collate_fn)
        
        if valid_data is not None:
            valid_loader = self._build_dataloader_from_dataset(valid_data, 
                num_workers=0, batch_size=self.batch_size, shuffle=False,
                collate_fn=trial_collate_fn)

        opt = torch.optim.Adam(self.model.parameters(), lr = self.lr, weight_decay = self.weight_decay)

        for epoch in range(self.epoch):
            for nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst in train_loader:
                features = []
                for smiles_lst in smiles_lst2: 
                    mol_lst = [Chem.MolFromSmiles(smiles) for smiles in smiles_lst]
                    mol_lst = list(filter(lambda x:x is not None, mol_lst))
                    fp_lst = [AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=512) for mol in mol_lst]
                    feature = np.zeros(512)
                    for fp in fp_lst:
                        arr = np.zeros((0,), dtype=np.int8)
                        DataStructs.ConvertToNumpyArray(fp,arr)
                        feature += arr 
                    feature = feature.reshape(1,-1)
                    features.append(feature)
                features = np.concatenate(features, 0)
                features = Variable(torch.from_numpy(features))
                y = self.model(features) 
                y = y.view(-1)
                loss = self.loss(y, label_vec.float())
                opt.zero_grad() 
                loss.backward() 
                opt.step()

            if valid_data is not None:
                eval_res = self._evaluate(valid_loader)
                for k, v in eval_res.items():
                    print('Epoch %d, %s: %.4f' % (epoch, k, v))
                
    def predict(self, test_data):
        '''
        Make clinical trial outcome predictions.
        
        Parameters
        ----------
        test_data: TrialOutcomeDatasetBase
            Testing data, should be a `TrialOutcomeDatasetBase` object.
        '''

        test_loader = self._build_dataloader_from_dataset(test_data, 
            num_workers=0, batch_size=self.batch_size, shuffle=False,
            collate_fn=trial_collate_fn)
        res = self._predict_on_dataloader(test_loader)
        nctids, ypred = res['nctid'], res['ypred']
        prediction = list(zip(nctids, ypred))
        return prediction

    def save_model(self, output_dir = None):
        '''
        Save the learned model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The output folder to save the learned model.
            If set None, will save model to `save_model/model.ckpt`.
        '''
        if output_dir is None:
            output_dir = 'save_model'
        if not os.path.exists(output_dir): os.makedirs(output_dir)
        filename = os.path.join(output_dir, 'model.ckpt')
        torch.save(self.model, filename)

        config_filename = os.path.join(output_dir, 'config.json')
        with open(config_filename, 'w') as f:
            json.dump(self.config, f)

    def load_model(self, checkpoint=None):
        '''
        Load the learned MLP model from the disk.

        Parameters
        ----------
        checkpoint: str 
            The checkpoint folder to load the learned model.
            The checkpoint under this folder should be `model.ckpt`.
        '''
        if checkpoint is None:
            ckpt_dir = 'save_model'
            checkpoint = os.path.join(ckpt_dir, 'model.ckpt')
        else:
            checkpoint = os.path.join(checkpoint, 'model.ckpt')
        self.model = torch.load(checkpoint)

        ckpt_dir = os.path.dirname(checkpoint)
        config_filename = os.path.join(ckpt_dir, 'config.json')
        with open(config_filename, 'r') as f:
            self.config = json.load(f)

    def _evaluate(self, dataloader):
        res = self._predict_on_dataloader(dataloader)
        nctids, ypred, ytrue = res['nctid'], res['ypred'], res['label']
        return {'auc': roc_auc_score(ytrue, ypred)}

    def _predict_on_dataloader(self, test_loader):
        nctids = []
        ypred = []
        labels = []
        for nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst in test_loader:
            features = []
            for smiles_lst in smiles_lst2: 
                mol_lst = [Chem.MolFromSmiles(smiles) for smiles in smiles_lst]
                mol_lst = list(filter(lambda x:x is not None, mol_lst))
                fp_lst = [AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=512) for mol in mol_lst]
                feature = np.zeros(512)
                for fp in fp_lst:
                    arr = np.zeros((0,), dtype=np.int8)
                    DataStructs.ConvertToNumpyArray(fp,arr)
                    feature += arr
                feature = feature.reshape(1,-1)
                features.append(feature)
            features = np.concatenate(features, 0)
            features = Variable(torch.from_numpy(features))
            y = self.model(features)
            y = torch.sigmoid(y)
            nctids.extend(nctid_lst)
            ypred.extend(y.tolist())
            if label_vec is not None:
                labels.extend(label_vec.tolist())
        res_dict = {'nctid': nctids, 'ypred': ypred, 'label': None}
        if len(labels) > 0:
            res_dict['label'] = labels
        return res_dict





# text descriptions of diagnosis codes are in 'examples/resources/CMS28_DESC_LONG_SHORT_DX.xlsx'
# text descriptions of procedure codes are in 'examples/resources/CMS28_DESC_LONG_SHORT_SG.xlsx'
# icd9 hierarchy is in 'examples/resources/icd-9-hierarchy.json'

from .models.base import PatientTrialMatchBase
from .models.compose import COMPOSE
from .models.deepenroll import DeepEnroll


###### import ######
import os
import pickle
import numpy as np
import torch
from torch.utils.data.dataloader import default_collate
from rdkit import Chem 
from rdkit.Chem import AllChem
from rdkit import RDLogger  
RDLogger.DisableLog('rdApp.info')
RDLogger.DisableLog('rdApp.*')
###### import ######



def plot_hist(prefix_name, prediction, label):
	import seaborn as sns
	import matplotlib.pyplot as plt
	figure_name = prefix_name + "_histogram.png"
	positive_prediction = [prediction[i] for i in range(len(label)) if label[i]==1]
	negative_prediction = [prediction[i] for i in range(len(label)) if label[i]==0]
	save_file_name = "results/" + prefix_name.split('/')[-1] + "_positive_negative.pkl"
	pickle.dump((positive_prediction, negative_prediction), open(save_file_name, 'wb'))
	# sns.distplot(positive_prediction, hist=True,  kde=False, bins=20, color = 'blue', label = 'success')  #### bins = 50 -> 20 
	# sns.distplot(negative_prediction, hist=True,  kde=False, bins=20, color = 'red', label = 'fail')
	# plt.xlabel("predicted success probability", fontsize=24)
	# plt.ylabel("frequencies", fontsize = 25)
	# plt.legend(fontsize = 21)
	# plt.tight_layout()
	# # plt.show()
	# plt.savefig(figure_name)
	return 

def replace_strange_symbol(text):
	for i in "[]'\n/":
		text = text.replace(i,'_')
	return text

#  xml read blog:  https://blog.csdn.net/yiluochenwu/article/details/23515923 
def walkData(root_node, prefix, result_list):
	temp_list =[prefix + '/' + root_node.tag, root_node.text]
	result_list.append(temp_list)
	children_node = root_node.getchildren()
	if len(children_node) == 0:
		return
	for child in children_node:
		walkData(child, prefix = prefix + '/' + root_node.tag, result_list = result_list)


def dynamic_programming(s1, s2):
	arr2d = [[0 for i in s2] for j in s1]
	if s1[0] == s2[0]:
		arr2d[0][0] = 1
	for i in range(1, len(s1)):
		if s1[i]==s2[0]:
			arr2d[i][0] = 1
		else:
			arr2d[i][0] = arr2d[i-1][0] 
	for i in range(1,len(s2)):
		if s2[i]==s1[0]:
			arr2d[0][i] = 1 
		else:
			arr2d[0][i] = arr2d[0][i-1]
	for i in range(1,len(s1)):
		for j in range(1,len(s2)):
			if s1[i] == s2[j]:
				arr2d[i][j] = arr2d[i-1][j-1] + 1 
			else:
				arr2d[i][j] = max(arr2d[i-1][j], arr2d[i][j-1])
	return arr2d[len(s1)-1][len(s2)-1]


def get_path_of_all_xml_file():
	input_file = "./data/all_xml"
	with open(input_file, 'r') as fin:
		lines = fin.readlines()
	input_file_lst = [i.strip() for i in lines]
	return input_file_lst 


def remove_multiple_space(text):
	text = ' '.join(text.split())
	return text 

def nctid_2_xml_file_path(nctid):
	assert len(nctid)==11
	prefix = nctid[:7] + "xxxx"
	datafolder = os.path.join("./ClinicalTrialGov/", prefix, nctid+".xml")
	return datafolder 


def fingerprints_from_mol(mol):
    fp = AllChem.GetMorganFingerprint(mol, 3, useCounts=True, useFeatures=True)
    size = 2048
    nfp = np.zeros((1, size), np.int32)
    for idx,v in fp.GetNonzeroElements().items():
        nidx = idx%size
        nfp[0, nidx] += int(v)
    return nfp

def smiles2fp(smiles):
	try:
		mol = Chem.MolFromSmiles(smiles)
		fp = fingerprints_from_mol(mol)
		return fp 
	except:
		return np.zeros((1, 2048), np.int32)

def smiles_lst2fp(smiles_lst):
	fp_lst = [smiles2fp(smiles) for smiles in smiles_lst]
	fp_mat = np.concatenate(fp_lst, 0)
	fp = np.mean(fp_mat,0)
	return fp	


#################  for data loader  #################
def clean_protocol(protocol):
	protocol = str(protocol).lower()
	protocol_split = protocol.split('\n')
	filter_out_empty_fn = lambda x: len(x.strip())>0
	strip_fn = lambda x:x.strip()
	protocol_split = list(filter(filter_out_empty_fn, protocol_split))	
	protocol_split = list(map(strip_fn, protocol_split))
	return protocol_split 

def split_protocol(protocol):
	protocol_split = clean_protocol(protocol)
	inclusion_idx, exclusion_idx = len(protocol_split), len(protocol_split)	
	for idx, sentence in enumerate(protocol_split):
		if "inclusion" in sentence:
			inclusion_idx = idx
			break
	for idx, sentence in enumerate(protocol_split):
		if "exclusion" in sentence:
			exclusion_idx = idx 
			break 		
	if inclusion_idx + 1 < exclusion_idx + 1 < len(protocol_split):
		inclusion_criteria = protocol_split[inclusion_idx:exclusion_idx]
		exclusion_criteria = protocol_split[exclusion_idx:]
		if not (len(inclusion_criteria) > 0 and len(exclusion_criteria) > 0):
			print(len(inclusion_criteria), len(exclusion_criteria), len(protocol_split))
			exit()
		return inclusion_criteria, exclusion_criteria ## list, list 
	else:
		return protocol_split, 

def protocol2feature(protocol, sentence_2_vec):
	result = split_protocol(protocol)
	inclusion_criteria, exclusion_criteria = result[0], result[-1]
	inclusion_feature = [sentence_2_vec[sentence].view(1,-1) for sentence in inclusion_criteria if sentence in sentence_2_vec]
	exclusion_feature = [sentence_2_vec[sentence].view(1,-1) for sentence in exclusion_criteria if sentence in sentence_2_vec]
	if inclusion_feature == []:
		inclusion_feature = torch.zeros(1,768)
	else:
		inclusion_feature = torch.cat(inclusion_feature, 0)
	if exclusion_feature == []:
		exclusion_feature = torch.zeros(1,768)
	else:
		exclusion_feature = torch.cat(exclusion_feature, 0)
	return inclusion_feature, exclusion_feature

def smiles_txt_to_lst(text):
	"""
		"['CN[C@H]1CC[C@@H](C2=CC(Cl)=C(Cl)C=C2)C2=CC=CC=C12', 
		  'CNCCC=C1C2=CC=CC=C2CCC2=CC=CC=C12']" 
	"""
	text = text[1:-1]
	lst = [i.strip()[1:-1] for i in text.split(',')]
	return lst 

def icdcode_text_2_lst_of_lst(text):
	text = text[2:-2]
	lst_lst = []
	for i in text.split('", "'):
		i = i[1:-1]
		lst_lst.append([j.strip()[1:-1] for j in i.split(',')])
	return lst_lst

def trial_collate_fn(x):
	sentence2vec = dict()
	nctid_lst = [i[0] for i in x]     ### ['NCT00604461', ..., 'NCT00788957'] 
	label_vec = default_collate([int(i[1]) for i in x])  ### shape n, 
	smiles_lst = [smiles_txt_to_lst(i[2]) for i in x]
	icdcode_lst = [icdcode_text_2_lst_of_lst(i[3]) for i in x]
	criteria_lst = [protocol2feature(i[4], sentence2vec) for i in x]
	return [nctid_lst, label_vec, smiles_lst, icdcode_lst, criteria_lst]


if __name__ == "__main__":
	text = "interpret_result/NCT00329602__completed____1__1.7650960683822632__phase 4__['restless legs syndrome']__['placebo', 'ropinirole'].png"
	print(replace_strange_symbol(text))






# if __name__ == "__main__":
# 	input_file_lst = get_path_of_all_xml_file() 
# 	print(input_file_lst[:5])
# '''
# input_file_lst = [ 
# 	'ClinicalTrialGov/NCT0000xxxx/NCT00000102.xml', 
#  	'ClinicalTrialGov/NCT0000xxxx/NCT00000104.xml', 
#  	'ClinicalTrialGov/NCT0000xxxx/NCT00000105.xml', 
# 	  ... ]
# '''



# if __name__ == "__main__":
# 	s1 = "328943"
# 	s2 = "13785"
# 	assert dynamic_programming(s1, s2)==2 








from sklearn.metrics import roc_auc_score, f1_score, average_precision_score, precision_score, recall_score, accuracy_score
import numpy as np
import pdb

def find_the_best_threshold(pred, label):
    best_threshold = 0
    best_f1 = 0
    for threshold in np.arange(0, 1, 0.1):
        pred_label = pred.copy()
        pred_label[pred_label >= threshold] = 1
        pred_label[pred_label < threshold] = 0
        f1score = f1_score(label, pred_label)
        if f1score > best_f1:
            best_f1 = f1score
            best_threshold = threshold
    return best_threshold

def bootstrap_evaluate(pred, label, sample_num=100):
    roc_auc_list, prauc_score_list, f1score_list = [], [], []
    for _ in range(sample_num):
        idx = np.random.choice(len(label), len(label), replace=True)
        pred_sample = pred[idx]
        label_sample = label[idx]
        res = evaluate(pred_sample, label_sample, find_best_threshold=True)
        roc_auc_list.append(res['roc_auc'])
        prauc_score_list.append(res['pr_auc'])
        f1score_list.append(res['f1'])
    print("PR-AUC   mean: "+str(np.mean(prauc_score_list))[:6], "std: "+str(np.std(prauc_score_list))[:6])
    print("F1       mean: "+str(np.mean(f1score_list))[:6], "std: "+str(np.std(f1score_list))[:6])
    print("ROC-AUC  mean: "+ str(np.mean(roc_auc_list))[:6], "std: " + str(np.std(roc_auc_list))[:6])
    return {'roc_auc': np.mean(roc_auc_list), 'pr_auc': np.mean(prauc_score_list), 'f1': np.mean(f1score_list)}

def evaluate(pred, label, find_best_threshold=False):
    if isinstance(pred, list):
        pred = np.array(pred)
    if isinstance(label, list):
        label = np.array(label)
    roc_auc = roc_auc_score(label, pred)
    prauc_score = average_precision_score(label, pred)
    pred_label = pred.copy()
    threshold = 0.5
    if find_best_threshold:
        threshold = find_the_best_threshold(pred, label)
    pred_label[pred_label >= threshold] = 1
    pred_label[pred_label < threshold] = 0
    f1score = f1_score(label, pred_label)
    # print('Evaluate on # of test trials: {}'.format(len(pred_label)))
    return {'pr_auc': prauc_score, 'f1': f1score, 'roc_auc': roc_auc}


'''
Provide a series of tools for ICD-9/10 codes.
(1) extract ICD codes given string terms.
(2) get parental and children nodes for ICD-9 codes.
'''
import requests
import os
import json
import wget
import pdb

import pandas as pd
import networkx as nx

NIH_API_PREFIX_ICD10 = 'https://clinicaltables.nlm.nih.gov/api/icd10cm/v3/search?sf=code,name&terms='
NIH_API_PREFIX_ICD9_DX = 'https://clinicaltables.nlm.nih.gov/api/icd9cm_dx/v3/search?sf=code,long_name&terms='
NIH_API_PREFIX_ICD9_SG = 'https://clinicaltables.nlm.nih.gov/api/icd9cm_sg/v3/search?sf=code,long_name&terms='
NIH_API_PREFIX_ICD9_CONDITION = 'https://clinicaltables.nlm.nih.gov/api/conditions/v3/search?df=primary_name&terms='
NIH_API_PREFIX_ICD10_CONDITION = 'https://clinicaltables.nlm.nih.gov/api/conditions/v3/search?df=term_icd10cm_codes,primary_name&terms='
ICD9_SG_URL = 'https://github.com/RyanWangZf/PyTrial/raw/main/resources/CMS32_DESC_LONG_SHORT_SG.xlsx'
ICD9_DX_URL = 'https://github.com/RyanWangZf/PyTrial/raw/main/resources/CMS32_DESC_LONG_SHORT_DX.xlsx'
ICD9_GRAPH_URL = 'https://storage.googleapis.com/pytrial/resources/icd-9-hierarchy.json'

def get_icd10_from_nih(term):
    '''
    Query related ICD-10 codes for input terms.

    Parameters
    ----------
    term: str or list[str]
        Disease names or a list of disease names.

    Returns
    -------
    Outputs ICD codes: list[str] or list[list[str]]
    '''
    if isinstance(term, str):
        terms = [term]
    else:
        terms = term

    outputs = []
    for term_ in terms:
        url = NIH_API_PREFIX_ICD10 + term_
        response = requests.get(url)
        text = response.text
        if text == '[0,[],null,[]]':
            outputs.append(None)
            continue
        text = text[1:-1]
        idx1 = text.find('[')
        idx2 = text.find(']')
        codes = text[idx1+1:idx2].split(',')
        codes = [i[1:-1] for i in codes]
        outputs.append(codes)

    if isinstance(term, str):
        outputs = outputs[0]

    return outputs

def get_icd9dx_from_nih(term):
    '''
    Query related ICD-9-CM diagnosis codes for input terms.

    Parameters
    ----------
    term: str or list[str]
        Disease names or a list of disease names.

    Returns
    -------
    Outputs ICD codes: list[str] or list[list[str]]
    '''
    if isinstance(term, str):
        terms = [term]
    else:
        terms = term

    outputs = []
    for term_ in terms:
        url = NIH_API_PREFIX_ICD9_DX + term_
        response = requests.get(url)
        text = response.text
        if text == '[0,[],null,[]]':
            outputs.append(None)
            continue
        text = text[1:-1]
        idx1 = text.find('[')
        idx2 = text.find(']')
        codes = text[idx1+1:idx2].split(',')
        codes = [i[1:-1] for i in codes]
        outputs.append(codes)

    if isinstance(term, str):
        outputs = outputs[0]

    return outputs

def get_icd9sg_from_nih(term):
    '''
    Query related ICD-9-CM procedure codes for input terms.

    Parameters
    ----------
    term: str or list[str]
        Disease names or a list of disease names.

    Returns
    -------
    Outputs ICD codes: list[str] or list[list[str]]
    '''
    if isinstance(term, str):
        terms = [term]
    else:
        terms = term

    outputs = []
    for term_ in terms:
        url = NIH_API_PREFIX_ICD9_SG + term_
        response = requests.get(url)
        text = response.text
        if text == '[0,[],null,[]]':
            outputs.append(None)
            continue
        text = text[1:-1]
        idx1 = text.find('[')
        idx2 = text.find(']')
        codes = text[idx1+1:idx2].split(',')
        codes = [i[1:-1] for i in codes]
        outputs.append(codes)

    if isinstance(term, str):
        outputs = outputs[0]

    return outputs

def get_condition_synonym_from_nih(term):
    '''
    Query relevant medical conditions taking input symptoms/diseases using API: https://clinicaltables.nlm.nih.gov/apidoc/conditions/v3/doc.html

    Parameters
    ----------
    term: str or list[str]
        Disease names or a list of disease names.

    Returns
    -------
    Outputs ICD codes: list[str] or list[list[str]]
    '''
    if isinstance(term, str):
        terms = [term]
    else:
        terms = term

    outputs = []
    for term_ in terms:
        url = NIH_API_PREFIX_ICD9_CONDITION + term_
        response = requests.get(url)
        text = response.text
        if text == '[0,[],null,[]]':
            outputs.append(None)
            continue
        text = text[1:-1]
        idx1 = text.find('[')
        idx2 = text.find(']')
        text = text[idx2+1:]
        idx1 = text.find('[')
        names = text[idx1+1:-1].split(',')
        names = [n[2:-2].lower() for n in names]
        outputs.append(names)

    if isinstance(term, str):
        outputs = outputs[0]

    return outputs

class ICDGraphBase:
    '''
    The base class for ICD9/10 graph.
    '''
    def __init__(self, filename):
        hierarchy = json.loads(open(filename, 'r').read())
        self.graph = nx.readwrite.json_graph.tree_graph(
            hierarchy['tree'],
            attrs={'id':'id', 'children':'children', 'description':'description'}
        )

    def children(self, code):
        '''Return children nodes of code.

        Returns
        -------
        children: list[dict]
            The children codes and their descriptions.
        '''
        code = code.replace('.','')
        node_list = list(self.graph.successors(code))
        return_list = []
        for node in node_list:
            return_list.append(self.__getitem__(node))
        return return_list

    def parent(self, code):
        '''Return the parent node of code.

        Returns
        -------
        parent: dict
            The parent code and its descriptions.
        '''
        code = code.replace('.','')
        node_list = list(self.graph.predecessors(code))
        return_list = []
        for node in node_list:
            return_list.append(self.__getitem__(node))
        return return_list[0]

    def siblings(self, code):
        '''Return sibling nodes of code.

        Returns
        -------
        siblings: list[dict]
            The sibling codes and their descriptions.
        '''
        code = code.replace('.', '')
        parentnode = self.parent(code)['code']
        return self.children(parentnode)

    @property
    def nodes(self):
        return self.graph.nodes

    @property
    def edges(self):
        return self.graph.edges

    @property
    def nxgraph(self):
        return self.graph

    def __getitem__(self, code):
        '''Return the description dict of the codes.
        '''
        code = code.replace('.','')
        return_dict = self.graph.nodes()[code]
        return_dict['code'] = code
        return return_dict

class ICD9Graph(ICDGraphBase):
    '''
    Get an ICD-9 knowledge graph to query parental and children nodes for each code.

    Returns
    -------
    self.graph: nx.DiGraph
        The hierarchy of ICD codes stored as graph in networkx.

    self.codes: list[str]
        All the unique codes.
    '''
    def __init__(self, input_dir=None):
        if input_dir is None:
            input_dir = './resources/icd9'
        
        if not os.path.exists(input_dir):
            os.makedirs(input_dir)
            # download the ICD9 hierarchy
            wget.download(ICD9_GRAPH_URL, input_dir)
        
        filename = os.path.join(input_dir, 'icd-9-hierarchy.json')
        super().__init__(filename)


class ICD10Graph(ICDGraphBase):
    '''
    Get an ICD-10 knowledge graph to query parental and children nodes for each code.

    Parameters
    ----------
    input_dir: str
        The dir that stores the hierarchy files.

    version: {'2022', '2021','2020','2019'}
        The version of ICD-10 codes.

    Returns
    -------
    self.graph: nx.DiGraph
        The hierarchy of ICD codes stored as graph in networkx.

    self.codes: list[str]
        All the unique codes.
    '''
    def __init__(self, input_dir=None, version='2021'):

        if input_dir is None:
            input_dir = './resources/icd10'

        if not os.path.exists(input_dir):
            os.makedirs(input_dir)
            # download the ICD10 hierarchy
            url = f'https://github.com/icd-codex/icd-codex/raw/dev/icdcodex/data/icd-10-{version}-hierarchy.json'
            wget.download(url, input_dir)

        filename = os.path.join(input_dir, f'icd-10-{version}-hierarchy.json')
        super().__init__(filename)

class ICD9_DX_VOC:
    '''
    Get a vocabulary containing the mapping of ICD9 Diagnosis code and its names.

    Parameters
    ----------
    input_dir: str
        The dir that stores ICD9-dx file.    
    '''
    def __init__(self, input_dir='./resources') -> None:
        url = ICD9_DX_URL
        filename = os.path.join(input_dir, 'CMS32_DESC_LONG_SHORT_DX.xlsx')
        if not os.path.exists(filename):
            # download to disk
            wget.download(url, out=filename)
        self.df = pd.read_excel(filename, dtype={'DIAGNOSIS CODE':str})
    
    def __getitem__(self, code):
        index = self.df['DIAGNOSIS CODE'].isin([code])
        sum_index = index.sum()
        if sum_index == 0:
            return None
        else:
            return self.df[index]['LONG DESCRIPTION'].tolist()[0]

    def code2desc(self, code):
        '''
        Get description of codes.

        Parameters
        ----------
        code: str or List[str]
            The input icd code or list of codes.
        '''
        if isinstance(code, str):
            code = [code]
        res = []
        for code_ in code:
            res.append(self.__getitem__(code_))
        if len(res) == 1:
            return res[0]
        else:
            return res

class ICD9_SG_VOC:
    '''
    Get a vocabulary containing the mapping of ICD9 procedure code and its names.

    Parameters
    ----------
    input_dir: str
        The dir that stores ICD9-sg file.
    '''
    def __init__(self, input_dir='./resources') -> None:
        url = ICD9_SG_URL
        filename = os.path.join(input_dir, 'CMS32_DESC_LONG_SHORT_SG.xlsx')
        if not os.path.exists(filename):
            # download to disk
            wget.download(url, out=filename)
        self.df = pd.read_excel(filename, dtype={'PROCEDURE CODE':str})

    def __getitem__(self, code):
        index = self.df['PROCEDURE CODE'].isin([code])
        sum_index = index.sum()
        if sum_index == 0:
            return None
        else:
            return self.df[index]['LONG DESCRIPTION'].tolist()[0]

    def code2desc(self, code):
        '''
        Get description of codes.

        Parameters
        ----------
        code: str or List[str]
            The input icd code or list of codes.
        '''
        if isinstance(code, str):
            code = [code]
        res = []
        for code_ in code:
            res.append(self.__getitem__(code_, 'LONG DESCRIPTION'))
        if len(res) == 1:
            return res[0]
        else:
            return res

if __name__ == '__main__':
    # graph = ICD9Graph('./resources')
    # graph = ICD10Graph('./resources')
    # print(get_icd10_from_nih(["lung neoplasm", "breast"]))
    # print(get_icd9dx_from_nih(["lung neoplasm", "breast"]))
    # print(get_icd9sg_from_nih(["lung neoplasm", "breast"]))
    # print(get_condition_synonym_from_nih(['gastroenteri','salmonella']))
    pass
